"""Builder class to transform a mypy AST to the IR form.

See the docstring of class IRBuilder for more information.
"""

from __future__ import annotations

from collections.abc import Iterator, Sequence
from contextlib import contextmanager
from typing import Any, Callable, Final, Union, overload

from mypy.build import Graph
from mypy.maptype import map_instance_to_supertype
from mypy.nodes import (
    ARG_NAMED,
    ARG_POS,
    GDEF,
    LDEF,
    PARAM_SPEC_KIND,
    TYPE_VAR_KIND,
    TYPE_VAR_TUPLE_KIND,
    ArgKind,
    CallExpr,
    Decorator,
    Expression,
    FuncDef,
    IndexExpr,
    IntExpr,
    Lvalue,
    MemberExpr,
    MypyFile,
    NameExpr,
    OpExpr,
    OverloadedFuncDef,
    RefExpr,
    StarExpr,
    Statement,
    SymbolNode,
    TupleExpr,
    TypeAlias,
    TypeInfo,
    TypeParam,
    UnaryExpr,
    Var,
)
from mypy.types import (
    AnyType,
    DeletedType,
    Instance,
    ProperType,
    TupleType,
    Type,
    TypedDictType,
    TypeOfAny,
    TypeVarLikeType,
    UninhabitedType,
    UnionType,
    get_proper_type,
)
from mypy.util import module_prefix, split_target
from mypy.visitor import ExpressionVisitor, StatementVisitor
from mypyc.common import BITMAP_BITS, SELF_NAME, TEMP_ATTR_NAME
from mypyc.crash import catch_errors
from mypyc.errors import Errors
from mypyc.ir.class_ir import ClassIR, NonExtClassInfo
from mypyc.ir.func_ir import INVALID_FUNC_DEF, FuncDecl, FuncIR, FuncSignature, RuntimeArg
from mypyc.ir.ops import (
    NAMESPACE_MODULE,
    NAMESPACE_TYPE_VAR,
    Assign,
    BasicBlock,
    Branch,
    ComparisonOp,
    GetAttr,
    InitStatic,
    Integer,
    IntOp,
    LoadStatic,
    Op,
    PrimitiveDescription,
    RaiseStandardError,
    Register,
    SetAttr,
    TupleGet,
    Unreachable,
    Value,
)
from mypyc.ir.rtypes import (
    RInstance,
    RTuple,
    RType,
    RUnion,
    bitmap_rprimitive,
    c_pyssize_t_rprimitive,
    dict_rprimitive,
    int_rprimitive,
    is_float_rprimitive,
    is_list_rprimitive,
    is_none_rprimitive,
    is_object_rprimitive,
    is_tagged,
    is_tuple_rprimitive,
    none_rprimitive,
    object_rprimitive,
    str_rprimitive,
)
from mypyc.irbuild.context import FuncInfo, ImplicitClass
from mypyc.irbuild.ll_builder import LowLevelIRBuilder
from mypyc.irbuild.mapper import Mapper
from mypyc.irbuild.nonlocalcontrol import (
    BaseNonlocalControl,
    GeneratorNonlocalControl,
    LoopNonlocalControl,
    NonlocalControl,
)
from mypyc.irbuild.prebuildvisitor import PreBuildVisitor
from mypyc.irbuild.prepare import RegisterImplInfo
from mypyc.irbuild.targets import (
    AssignmentTarget,
    AssignmentTargetAttr,
    AssignmentTargetIndex,
    AssignmentTargetRegister,
    AssignmentTargetTuple,
)
from mypyc.irbuild.util import bytes_from_str, is_constant
from mypyc.options import CompilerOptions
from mypyc.primitives.dict_ops import dict_get_item_op, dict_set_item_op
from mypyc.primitives.generic_ops import iter_op, next_op, py_setattr_op
from mypyc.primitives.list_ops import list_get_item_unsafe_op, list_pop_last, to_list
from mypyc.primitives.misc_ops import check_unpack_count_op, get_module_dict_op, import_op
from mypyc.primitives.registry import CFunctionDescription, function_ops

# These int binary operations can borrow their operands safely, since the
# primitives take this into consideration.
int_borrow_friendly_op: Final = {"+", "-", "==", "!=", "<", "<=", ">", ">="}


class IRVisitor(ExpressionVisitor[Value], StatementVisitor[None]):
    pass


class UnsupportedException(Exception):
    pass


SymbolTarget = Union[AssignmentTargetRegister, AssignmentTargetAttr]


class IRBuilder:
    """Builder class used to construct mypyc IR from a mypy AST.

    The IRBuilder class maintains IR transformation state and provides access
    to various helpers used to implement the transform.

    mypyc.irbuild.visitor.IRBuilderVisitor is used to dispatch based on mypy
    AST node type to code that actually does the bulk of the work. For
    example, expressions are transformed in mypyc.irbuild.expression and
    functions are transformed in mypyc.irbuild.function.

    Use the "accept()" method to translate individual mypy AST nodes to IR.
    Other methods are used to generate IR for various lower-level operations.

    This class wraps the lower-level LowLevelIRBuilder class, an instance
    of which is available through the "builder" attribute. The low-level
    builder class doesn't have any knowledge of the mypy AST. Wrappers for
    some LowLevelIRBuilder method are provided for convenience, but others
    can also be accessed via the "builder" attribute.

    See also:
     * The mypyc IR is defined in the mypyc.ir package.
     * The top-level IR transform control logic is in mypyc.irbuild.main.
    """

    def __init__(
        self,
        current_module: str,
        types: dict[Expression, Type],
        graph: Graph,
        errors: Errors,
        mapper: Mapper,
        pbv: PreBuildVisitor,
        visitor: IRVisitor,
        options: CompilerOptions,
        singledispatch_impls: dict[FuncDef, list[RegisterImplInfo]],
    ) -> None:
        self.builder = LowLevelIRBuilder(errors, options)
        self.builders = [self.builder]
        self.symtables: list[dict[SymbolNode, SymbolTarget]] = [{}]
        self.runtime_args: list[list[RuntimeArg]] = [[]]
        self.function_name_stack: list[str] = []
        self.class_ir_stack: list[ClassIR] = []
        # Keep track of whether the next statement in a block is reachable
        # or not, separately for each block nesting level
        self.block_reachable_stack: list[bool] = [True]

        self.current_module = current_module
        self.mapper = mapper
        self.types = types
        self.graph = graph
        self.ret_types: list[RType] = []
        self.functions: list[FuncIR] = []
        self.function_names: set[tuple[str | None, str]] = set()
        self.classes: list[ClassIR] = []
        self.final_names: list[tuple[str, RType]] = []
        self.type_var_names: list[str] = []
        self.callable_class_names: set[str] = set()
        self.options = options

        # These variables keep track of the number of lambdas, implicit indices, and implicit
        # iterators instantiated so we avoid name conflicts. The indices and iterators are
        # instantiated from for-loops.
        self.lambda_counter = 0
        self.temp_counter = 0

        # These variables are populated from the first-pass PreBuildVisitor.
        self.free_variables = pbv.free_variables
        self.prop_setters = pbv.prop_setters
        self.encapsulating_funcs = pbv.encapsulating_funcs
        self.nested_fitems = pbv.nested_funcs.keys()
        self.fdefs_to_decorators = pbv.funcs_to_decorators
        self.module_import_groups = pbv.module_import_groups

        self.singledispatch_impls = singledispatch_impls

        self.visitor = visitor

        # This list operates similarly to a function call stack for nested functions. Whenever a
        # function definition begins to be generated, a FuncInfo instance is added to the stack,
        # and information about that function (e.g. whether it is nested, its environment class to
        # be generated) is stored in that FuncInfo instance. When the function is done being
        # generated, its corresponding FuncInfo is popped off the stack.
        self.fn_info = FuncInfo(INVALID_FUNC_DEF, "", "")
        self.fn_infos: list[FuncInfo] = [self.fn_info]

        # This list operates as a stack of constructs that modify the
        # behavior of nonlocal control flow constructs.
        self.nonlocal_control: list[NonlocalControl] = []

        self.errors = errors
        # Notionally a list of all of the modules imported by the
        # module being compiled, but stored as an OrderedDict so we
        # can also do quick lookups.
        self.imports: dict[str, None] = {}

        self.can_borrow = False

    # High-level control

    def set_module(self, module_name: str, module_path: str) -> None:
        """Set the name and path of the current module.

        This must be called before transforming any AST nodes.
        """
        self.module_name = module_name
        self.module_path = module_path
        self.builder.set_module(module_name, module_path)

    @overload
    def accept(self, node: Expression, *, can_borrow: bool = False) -> Value: ...

    @overload
    def accept(self, node: Statement) -> None: ...

    def accept(self, node: Statement | Expression, *, can_borrow: bool = False) -> Value | None:
        """Transform an expression or a statement.

        If can_borrow is true, prefer to generate a borrowed reference.
        Borrowed references are faster since they don't require reference count
        manipulation, but they are only safe to use in specific contexts.
        """
        with self.catch_errors(node.line):
            if isinstance(node, Expression):
                old_can_borrow = self.can_borrow
                self.can_borrow = can_borrow
                try:
                    res = node.accept(self.visitor)
                    res = self.coerce(res, self.node_type(node), node.line)
                # If we hit an error during compilation, we want to
                # keep trying, so we can produce more error
                # messages. Generate a temp of the right type to keep
                # from causing more downstream trouble.
                except UnsupportedException:
                    res = Register(self.node_type(node))
                self.can_borrow = old_can_borrow
                if not can_borrow:
                    self.flush_keep_alives()
                return res
            else:
                try:
                    node.accept(self.visitor)
                except UnsupportedException:
                    pass
                return None

    def flush_keep_alives(self) -> None:
        self.builder.flush_keep_alives()

    # Pass through methods for the most common low-level builder ops, for convenience.

    def add(self, op: Op) -> Value:
        return self.builder.add(op)

    def goto(self, target: BasicBlock) -> None:
        self.builder.goto(target)

    def activate_block(self, block: BasicBlock) -> None:
        self.builder.activate_block(block)

    def goto_and_activate(self, block: BasicBlock) -> None:
        self.builder.goto_and_activate(block)

    def self(self) -> Register:
        return self.builder.self()

    def py_get_attr(self, obj: Value, attr: str, line: int) -> Value:
        return self.builder.py_get_attr(obj, attr, line)

    def load_str(self, value: str) -> Value:
        return self.builder.load_str(value)

    def load_bytes_from_str_literal(self, value: str) -> Value:
        """Load bytes object from a string literal.

        The literal characters of BytesExpr (the characters inside b'')
        are stored in BytesExpr.value, whose type is 'str' not 'bytes'.
        Thus we perform a special conversion here.
        """
        return self.builder.load_bytes(bytes_from_str(value))

    def load_int(self, value: int) -> Value:
        return self.builder.load_int(value)

    def load_float(self, value: float) -> Value:
        return self.builder.load_float(value)

    def unary_op(self, lreg: Value, expr_op: str, line: int) -> Value:
        return self.builder.unary_op(lreg, expr_op, line)

    def binary_op(self, lreg: Value, rreg: Value, expr_op: str, line: int) -> Value:
        return self.builder.binary_op(lreg, rreg, expr_op, line)

    def coerce(self, src: Value, target_type: RType, line: int, force: bool = False) -> Value:
        return self.builder.coerce(src, target_type, line, force, can_borrow=self.can_borrow)

    def none_object(self) -> Value:
        return self.builder.none_object()

    def none(self) -> Value:
        return self.builder.none()

    def true(self) -> Value:
        return self.builder.true()

    def false(self) -> Value:
        return self.builder.false()

    def new_list_op(self, values: list[Value], line: int) -> Value:
        return self.builder.new_list_op(values, line)

    def new_set_op(self, values: list[Value], line: int) -> Value:
        return self.builder.new_set_op(values, line)

    def translate_is_op(self, lreg: Value, rreg: Value, expr_op: str, line: int) -> Value:
        return self.builder.translate_is_op(lreg, rreg, expr_op, line)

    def py_call(
        self,
        function: Value,
        arg_values: list[Value],
        line: int,
        arg_kinds: list[ArgKind] | None = None,
        arg_names: Sequence[str | None] | None = None,
    ) -> Value:
        return self.builder.py_call(function, arg_values, line, arg_kinds, arg_names)

    def add_bool_branch(self, value: Value, true: BasicBlock, false: BasicBlock) -> None:
        self.builder.add_bool_branch(value, true, false)

    def load_native_type_object(self, fullname: str) -> Value:
        return self.builder.load_native_type_object(fullname)

    def gen_method_call(
        self,
        base: Value,
        name: str,
        arg_values: list[Value],
        result_type: RType | None,
        line: int,
        arg_kinds: list[ArgKind] | None = None,
        arg_names: list[str | None] | None = None,
    ) -> Value:
        return self.builder.gen_method_call(
            base, name, arg_values, result_type, line, arg_kinds, arg_names, self.can_borrow
        )

    def load_module(self, name: str) -> Value:
        return self.builder.load_module(name)

    def call_c(self, desc: CFunctionDescription, args: list[Value], line: int) -> Value:
        return self.builder.call_c(desc, args, line)

    def primitive_op(
        self,
        desc: PrimitiveDescription,
        args: list[Value],
        line: int,
        result_type: RType | None = None,
    ) -> Value:
        return self.builder.primitive_op(desc, args, line, result_type)

    def int_op(self, type: RType, lhs: Value, rhs: Value, op: int, line: int) -> Value:
        return self.builder.int_op(type, lhs, rhs, op, line)

    def compare_tuples(self, lhs: Value, rhs: Value, op: str, line: int) -> Value:
        return self.builder.compare_tuples(lhs, rhs, op, line)

    def builtin_len(self, val: Value, line: int) -> Value:
        return self.builder.builtin_len(val, line)

    def new_tuple(self, items: list[Value], line: int) -> Value:
        return self.builder.new_tuple(items, line)

    def debug_print(self, toprint: str | Value) -> None:
        return self.builder.debug_print(toprint)

    # Helpers for IR building

    def add_to_non_ext_dict(
        self, non_ext: NonExtClassInfo, key: str, val: Value, line: int
    ) -> None:
        # Add an attribute entry into the class dict of a non-extension class.
        key_unicode = self.load_str(key)
        self.primitive_op(dict_set_item_op, [non_ext.dict, key_unicode, val], line)

    def gen_import(self, id: str, line: int) -> None:
        self.imports[id] = None

        needs_import, out = BasicBlock(), BasicBlock()
        self.check_if_module_loaded(id, line, needs_import, out)

        self.activate_block(needs_import)
        value = self.call_c(import_op, [self.load_str(id)], line)
        self.add(InitStatic(value, id, namespace=NAMESPACE_MODULE))
        self.goto_and_activate(out)

    def check_if_module_loaded(
        self, id: str, line: int, needs_import: BasicBlock, out: BasicBlock
    ) -> None:
        """Generate code that checks if the module `id` has been loaded yet.

        Arguments:
            id: name of module to check if imported
            line: line number that the import occurs on
            needs_import: the BasicBlock that is run if the module has not been loaded yet
            out: the BasicBlock that is run if the module has already been loaded"""
        first_load = self.load_module(id)
        comparison = self.translate_is_op(first_load, self.none_object(), "is not", line)
        self.add_bool_branch(comparison, out, needs_import)

    def get_module(self, module: str, line: int) -> Value:
        # Python 3.7 has a nice 'PyImport_GetModule' function that we can't use :(
        mod_dict = self.call_c(get_module_dict_op, [], line)
        # Get module object from modules dict.
        return self.primitive_op(dict_get_item_op, [mod_dict, self.load_str(module)], line)

    def get_module_attr(self, module: str, attr: str, line: int) -> Value:
        """Look up an attribute of a module without storing it in the local namespace.

        For example, get_module_attr('typing', 'TypedDict', line) results in
        the value of 'typing.TypedDict'.

        Import the module if needed.
        """
        self.gen_import(module, line)
        module_obj = self.get_module(module, line)
        return self.py_get_attr(module_obj, attr, line)

    def assign_if_null(self, target: Register, get_val: Callable[[], Value], line: int) -> None:
        """If target is NULL, assign value produced by get_val to it."""
        error_block, body_block = BasicBlock(), BasicBlock()
        self.add(Branch(target, error_block, body_block, Branch.IS_ERROR))
        self.activate_block(error_block)
        self.add(Assign(target, self.coerce(get_val(), target.type, line)))
        self.goto(body_block)
        self.activate_block(body_block)

    def assign_if_bitmap_unset(
        self, target: Register, get_val: Callable[[], Value], index: int, line: int
    ) -> None:
        error_block, body_block = BasicBlock(), BasicBlock()
        o = self.int_op(
            bitmap_rprimitive,
            self.builder.args[-1 - index // BITMAP_BITS],
            Integer(1 << (index & (BITMAP_BITS - 1)), bitmap_rprimitive),
            IntOp.AND,
            line,
        )
        b = self.add(ComparisonOp(o, Integer(0, bitmap_rprimitive), ComparisonOp.EQ))
        self.add(Branch(b, error_block, body_block, Branch.BOOL))
        self.activate_block(error_block)
        self.add(Assign(target, self.coerce(get_val(), target.type, line)))
        self.goto(body_block)
        self.activate_block(body_block)

    def maybe_add_implicit_return(self) -> None:
        if is_none_rprimitive(self.ret_types[-1]) or is_object_rprimitive(self.ret_types[-1]):
            self.add_implicit_return()
        else:
            self.add_implicit_unreachable()

    def add_implicit_return(self) -> None:
        block = self.builder.blocks[-1]
        if not block.terminated:
            retval = self.coerce(self.builder.none(), self.ret_types[-1], -1)
            self.nonlocal_control[-1].gen_return(self, retval, self.fn_info.fitem.line)

    def add_implicit_unreachable(self) -> None:
        block = self.builder.blocks[-1]
        if not block.terminated:
            self.add(Unreachable())

    def disallow_class_assignments(self, lvalues: list[Lvalue], line: int) -> None:
        # Some best-effort attempts to disallow assigning to class
        # variables that aren't marked ClassVar, since we blatantly
        # miscompile the interaction between instance and class
        # variables.
        for lvalue in lvalues:
            if (
                isinstance(lvalue, MemberExpr)
                and isinstance(lvalue.expr, RefExpr)
                and isinstance(lvalue.expr.node, TypeInfo)
            ):
                var = lvalue.expr.node[lvalue.name].node
                if isinstance(var, Var) and not var.is_classvar:
                    self.error("Only class variables defined as ClassVar can be assigned to", line)

    def non_function_scope(self) -> bool:
        # Currently the stack always has at least two items: dummy and top-level.
        return len(self.fn_infos) <= 2

    def top_level_fn_info(self) -> FuncInfo | None:
        if self.non_function_scope():
            return None
        return self.fn_infos[2]

    def init_final_static(
        self,
        lvalue: Lvalue,
        rvalue_reg: Value,
        class_name: str | None = None,
        *,
        type_override: RType | None = None,
    ) -> None:
        assert isinstance(lvalue, NameExpr)
        assert isinstance(lvalue.node, Var)
        if lvalue.node.final_value is None:
            if class_name is None:
                name = lvalue.name
            else:
                name = f"{class_name}.{lvalue.name}"
            assert name is not None, "Full name not set for variable"
            coerced = self.coerce(rvalue_reg, type_override or self.node_type(lvalue), lvalue.line)
            self.final_names.append((name, coerced.type))
            self.add(InitStatic(coerced, name, self.module_name))

    def load_final_static(
        self, fullname: str, typ: RType, line: int, error_name: str | None = None
    ) -> Value:
        split_name = split_target(self.graph, fullname)
        assert split_name is not None
        module, name = split_name
        return self.builder.load_static_checked(
            typ,
            name,
            module,
            line=line,
            error_msg=f'value for final name "{error_name}" was not set',
        )

    def init_type_var(self, value: Value, name: str, line: int) -> None:
        unique_name = name + "___" + str(line)
        self.type_var_names.append(unique_name)
        self.add(InitStatic(value, unique_name, self.module_name, namespace=NAMESPACE_TYPE_VAR))

    def load_type_var(self, name: str, line: int) -> Value:
        return self.add(
            LoadStatic(
                object_rprimitive,
                name + "___" + str(line),
                self.module_name,
                namespace=NAMESPACE_TYPE_VAR,
            )
        )

    def load_literal_value(self, val: int | str | bytes | float | complex | bool) -> Value:
        """Load value of a final name, class-level attribute, or constant folded expression."""
        if isinstance(val, bool):
            if val:
                return self.true()
            else:
                return self.false()
        elif isinstance(val, int):
            return self.builder.load_int(val)
        elif isinstance(val, float):
            return self.builder.load_float(val)
        elif isinstance(val, str):
            return self.builder.load_str(val)
        elif isinstance(val, bytes):
            return self.builder.load_bytes(val)
        elif isinstance(val, complex):
            return self.builder.load_complex(val)
        else:
            assert False, "Unsupported literal value"

    def get_assignment_target(
        self, lvalue: Lvalue, line: int = -1, *, for_read: bool = False
    ) -> AssignmentTarget:
        if line == -1:
            line = lvalue.line
        if isinstance(lvalue, NameExpr):
            # If we are visiting a decorator, then the SymbolNode we really want to be looking at
            # is the function that is decorated, not the entire Decorator node itself.
            symbol = lvalue.node
            if isinstance(symbol, Decorator):
                symbol = symbol.func
            if symbol is None:
                # Semantic analyzer doesn't create ad-hoc Vars for special forms.
                assert lvalue.is_special_form
                symbol = Var(lvalue.name)
            if not for_read and isinstance(symbol, Var) and symbol.is_cls:
                self.error("Cannot assign to the first argument of classmethod", line)
            if lvalue.kind == LDEF:
                if symbol not in self.symtables[-1]:
                    if isinstance(symbol, Var) and not isinstance(symbol.type, DeletedType):
                        reg_type = self.type_to_rtype(symbol.type)
                    else:
                        reg_type = self.node_type(lvalue)
                    # If the function is a generator function, then first define a new variable
                    # in the current function's environment class. Next, define a target that
                    # refers to the newly defined variable in that environment class. Add the
                    # target to the table containing class environment variables, as well as the
                    # current environment.
                    if self.fn_info.is_generator:
                        return self.add_var_to_env_class(
                            symbol, reg_type, self.fn_info.generator_class, reassign=False
                        )

                    # Otherwise define a new local variable.
                    return self.add_local_reg(symbol, reg_type)
                else:
                    # Assign to a previously defined variable.
                    return self.lookup(symbol)
            elif lvalue.kind == GDEF:
                globals_dict = self.load_globals_dict()
                name = self.load_str(lvalue.name)
                return AssignmentTargetIndex(globals_dict, name)
            else:
                assert False, lvalue.kind
        elif isinstance(lvalue, IndexExpr):
            # Indexed assignment x[y] = e
            base = self.accept(lvalue.base)
            index = self.accept(lvalue.index)
            return AssignmentTargetIndex(base, index)
        elif isinstance(lvalue, MemberExpr):
            # Attribute assignment x.y = e
            can_borrow = self.is_native_attr_ref(lvalue)
            obj = self.accept(lvalue.expr, can_borrow=can_borrow)
            return AssignmentTargetAttr(obj, lvalue.name, can_borrow=can_borrow)
        elif isinstance(lvalue, TupleExpr):
            # Multiple assignment a, ..., b = e
            star_idx: int | None = None
            lvalues = []
            for idx, item in enumerate(lvalue.items):
                targ = self.get_assignment_target(item)
                lvalues.append(targ)
                if isinstance(item, StarExpr):
                    if star_idx is not None:
                        self.error("Two starred expressions in assignment", line)
                    star_idx = idx

            return AssignmentTargetTuple(lvalues, star_idx)

        elif isinstance(lvalue, StarExpr):
            return self.get_assignment_target(lvalue.expr)

        assert False, "Unsupported lvalue: %r" % lvalue

    def read(
        self, target: Value | AssignmentTarget, line: int = -1, can_borrow: bool = False
    ) -> Value:
        if isinstance(target, Value):
            return target
        if isinstance(target, AssignmentTargetRegister):
            return target.register
        if isinstance(target, AssignmentTargetIndex):
            reg = self.gen_method_call(
                target.base, "__getitem__", [target.index], target.type, line
            )
            if reg is not None:
                return reg
            assert False, target.base.type
        if isinstance(target, AssignmentTargetAttr):
            if isinstance(target.obj.type, RInstance) and target.obj.type.class_ir.is_ext_class:
                borrow = can_borrow and target.can_borrow
                return self.add(GetAttr(target.obj, target.attr, line, borrow=borrow))
            else:
                return self.py_get_attr(target.obj, target.attr, line)

        assert False, "Unsupported lvalue: %r" % target

    def read_nullable_attr(self, obj: Value, attr: str, line: int = -1) -> Value:
        """Read an attribute that might be NULL without raising AttributeError.

        This is used for reading spill targets in try/finally blocks where NULL
        indicates the non-return path was taken.
        """
        assert isinstance(obj.type, RInstance) and obj.type.class_ir.is_ext_class
        return self.add(GetAttr(obj, attr, line, allow_null=True))

    def assign(self, target: Register | AssignmentTarget, rvalue_reg: Value, line: int) -> None:
        if isinstance(target, Register):
            self.add(Assign(target, self.coerce_rvalue(rvalue_reg, target.type, line)))
        elif isinstance(target, AssignmentTargetRegister):
            rvalue_reg = self.coerce_rvalue(rvalue_reg, target.type, line)
            self.add(Assign(target.register, rvalue_reg))
        elif isinstance(target, AssignmentTargetAttr):
            if isinstance(target.obj_type, RInstance):
                rvalue_reg = self.coerce_rvalue(rvalue_reg, target.type, line)
                self.add(SetAttr(target.obj, target.attr, rvalue_reg, line))
            else:
                key = self.load_str(target.attr)
                boxed_reg = self.builder.box(rvalue_reg)
                self.primitive_op(py_setattr_op, [target.obj, key, boxed_reg], line)
        elif isinstance(target, AssignmentTargetIndex):
            target_reg2 = self.gen_method_call(
                target.base, "__setitem__", [target.index, rvalue_reg], None, line
            )
            assert target_reg2 is not None, target.base.type
        elif isinstance(target, AssignmentTargetTuple):
            if isinstance(rvalue_reg.type, RTuple) and target.star_idx is None:
                rtypes = rvalue_reg.type.types
                assert len(rtypes) == len(target.items)
                for i in range(len(rtypes)):
                    item_value = self.add(TupleGet(rvalue_reg, i, line))
                    self.assign(target.items[i], item_value, line)
            elif (
                is_list_rprimitive(rvalue_reg.type) or is_tuple_rprimitive(rvalue_reg.type)
            ) and target.star_idx is None:
                self.process_sequence_assignment(target, rvalue_reg, line)
            else:
                self.process_iterator_tuple_assignment(target, rvalue_reg, line)
        else:
            assert False, "Unsupported assignment target"

    def coerce_rvalue(self, rvalue: Value, rtype: RType, line: int) -> Value:
        if is_float_rprimitive(rtype) and is_tagged(rvalue.type):
            typename = rvalue.type.short_name()
            if typename == "short_int":
                typename = "int"
            self.error(
                "Incompatible value representations in assignment "
                + f'(expression has type "{typename}", variable has type "float")',
                line,
            )
        return self.coerce(rvalue, rtype, line)

    def process_sequence_assignment(
        self, target: AssignmentTargetTuple, rvalue: Value, line: int
    ) -> None:
        """Process assignment like 'x, y = s', where s is a variable-length list or tuple."""
        # Check the length of sequence.
        expected_len = Integer(len(target.items), c_pyssize_t_rprimitive)
        self.builder.call_c(check_unpack_count_op, [rvalue, expected_len], line)

        # Read sequence items.
        values = []
        for i in range(len(target.items)):
            item = target.items[i]
            index = self.builder.load_int(i)
            if is_list_rprimitive(rvalue.type):
                item_value = self.primitive_op(list_get_item_unsafe_op, [rvalue, index], line)
            else:
                item_value = self.builder.gen_method_call(
                    rvalue, "__getitem__", [index], item.type, line
                )
            values.append(item_value)

        # Assign sequence items to the target lvalues.
        for lvalue, value in zip(target.items, values):
            self.assign(lvalue, value, line)

    def process_iterator_tuple_assignment_helper(
        self, litem: AssignmentTarget, ritem: Value, line: int
    ) -> None:
        error_block, ok_block = BasicBlock(), BasicBlock()
        self.add(Branch(ritem, error_block, ok_block, Branch.IS_ERROR))

        self.activate_block(error_block)
        self.add(
            RaiseStandardError(RaiseStandardError.VALUE_ERROR, "not enough values to unpack", line)
        )
        self.add(Unreachable())

        self.activate_block(ok_block)
        self.assign(litem, ritem, line)

    def process_iterator_tuple_assignment(
        self, target: AssignmentTargetTuple, rvalue_reg: Value, line: int
    ) -> None:
        iterator = self.primitive_op(iter_op, [rvalue_reg], line)

        # This may be the whole lvalue list if there is no starred value
        split_idx = target.star_idx if target.star_idx is not None else len(target.items)

        # Assign values before the first starred value
        for litem in target.items[:split_idx]:
            ritem = self.call_c(next_op, [iterator], line)
            error_block, ok_block = BasicBlock(), BasicBlock()
            self.add(Branch(ritem, error_block, ok_block, Branch.IS_ERROR))

            self.activate_block(error_block)
            self.add(
                RaiseStandardError(
                    RaiseStandardError.VALUE_ERROR, "not enough values to unpack", line
                )
            )
            self.add(Unreachable())

            self.activate_block(ok_block)

            self.assign(litem, ritem, line)

        # Assign the starred value and all values after it
        if target.star_idx is not None:
            post_star_vals = target.items[split_idx + 1 :]
            iter_list = self.primitive_op(to_list, [iterator], line)
            iter_list_len = self.builtin_len(iter_list, line)
            post_star_len = Integer(len(post_star_vals))
            condition = self.binary_op(post_star_len, iter_list_len, "<=", line)

            error_block, ok_block = BasicBlock(), BasicBlock()
            self.add(Branch(condition, ok_block, error_block, Branch.BOOL))

            self.activate_block(error_block)
            self.add(
                RaiseStandardError(
                    RaiseStandardError.VALUE_ERROR, "not enough values to unpack", line
                )
            )
            self.add(Unreachable())

            self.activate_block(ok_block)

            for litem in reversed(post_star_vals):
                ritem = self.primitive_op(list_pop_last, [iter_list], line)
                self.assign(litem, ritem, line)

            # Assign the starred value
            self.assign(target.items[target.star_idx], iter_list, line)

        # There is no starred value, so check if there are extra values in rhs that
        # have not been assigned.
        else:
            extra = self.call_c(next_op, [iterator], line)
            error_block, ok_block = BasicBlock(), BasicBlock()
            self.add(Branch(extra, ok_block, error_block, Branch.IS_ERROR))

            self.activate_block(error_block)
            self.add(
                RaiseStandardError(
                    RaiseStandardError.VALUE_ERROR, "too many values to unpack", line
                )
            )
            self.add(Unreachable())

            self.activate_block(ok_block)

    def push_loop_stack(self, continue_block: BasicBlock, break_block: BasicBlock) -> None:
        self.nonlocal_control.append(
            LoopNonlocalControl(self.nonlocal_control[-1], continue_block, break_block)
        )

    def pop_loop_stack(self) -> None:
        self.nonlocal_control.pop()

    def make_spill_target(self, type: RType) -> AssignmentTarget:
        """Moves a given Value instance into the generator class' environment class."""
        name = f"{TEMP_ATTR_NAME}{self.temp_counter}"
        self.temp_counter += 1
        target = self.add_var_to_env_class(Var(name), type, self.fn_info.generator_class)
        return target

    def spill(self, value: Value) -> AssignmentTarget:
        """Moves a given Value instance into the generator class' environment class."""
        target = self.make_spill_target(value.type)
        # Shouldn't be able to fail, so -1 for line
        self.assign(target, value, -1)
        return target

    def maybe_spill(self, value: Value) -> Value | AssignmentTarget:
        """
        Moves a given Value instance into the environment class for generator functions. For
        non-generator functions, leaves the Value instance as it is.

        Returns an AssignmentTarget associated with the Value for generator functions and the
        original Value itself for non-generator functions.
        """
        if self.fn_info.is_generator:
            return self.spill(value)
        return value

    def maybe_spill_assignable(self, value: Value) -> Register | AssignmentTarget:
        """
        Moves a given Value instance into the environment class for generator functions. For
        non-generator functions, allocate a temporary Register.

        Returns an AssignmentTarget associated with the Value for generator functions and an
        assignable Register for non-generator functions.
        """
        if self.fn_info.is_generator:
            return self.spill(value)

        if isinstance(value, Register):
            return value

        # Allocate a temporary register for the assignable value.
        reg = Register(value.type)
        self.assign(reg, value, -1)
        return reg

    def extract_int(self, e: Expression) -> int | None:
        if isinstance(e, IntExpr):
            return e.value
        elif isinstance(e, UnaryExpr) and e.op == "-" and isinstance(e.expr, IntExpr):
            return -e.expr.value
        else:
            return None

    def get_sequence_type(self, expr: Expression) -> RType:
        return self.get_sequence_type_from_type(self.types[expr])

    def get_sequence_type_from_type(self, target_type: Type) -> RType:
        target_type = get_proper_type(target_type)
        if isinstance(target_type, UnionType):
            return RUnion.make_simplified_union(
                [self.get_sequence_type_from_type(item) for item in target_type.items]
            )
        elif isinstance(target_type, Instance):
            if target_type.type.fullname == "builtins.str":
                return str_rprimitive
            else:
                return self.type_to_rtype(target_type.args[0])
        # This elif-blocks are needed for iterating over classes derived from NamedTuple.
        elif isinstance(target_type, TypeVarLikeType):
            return self.get_sequence_type_from_type(target_type.upper_bound)
        elif isinstance(target_type, TupleType):
            # Tuple might have elements of different types.
            rtypes = {self.mapper.type_to_rtype(item) for item in target_type.items}
            if len(rtypes) == 1:
                return rtypes.pop()
            else:
                return RUnion.make_simplified_union(list(rtypes))
        assert False, target_type

    def get_dict_base_type(self, expr: Expression) -> list[Instance]:
        """Find dict type of a dict-like expression.

        This is useful for dict subclasses like SymbolTable.
        """
        return self.get_dict_base_type_from_type(self.types[expr])

    def get_dict_base_type_from_type(self, target_type: Type) -> list[Instance]:
        target_type = get_proper_type(target_type)
        if isinstance(target_type, UnionType):
            return [
                inner
                for item in target_type.items
                for inner in self.get_dict_base_type_from_type(item)
            ]
        if isinstance(target_type, TypeVarLikeType):
            # Match behaviour of self.node_type
            # We can only reach this point if `target_type` was a TypeVar(bound=dict[...])
            # or a ParamSpec.
            return self.get_dict_base_type_from_type(target_type.upper_bound)

        if isinstance(target_type, TypedDictType):
            target_type = target_type.fallback
            dict_base = next(
                base for base in target_type.type.mro if base.fullname == "typing.Mapping"
            )
        elif isinstance(target_type, Instance):
            dict_base = next(
                base for base in target_type.type.mro if base.fullname == "builtins.dict"
            )
        else:
            assert False, f"Failed to extract dict base from {target_type}"
        return [map_instance_to_supertype(target_type, dict_base)]

    def get_dict_key_type(self, expr: Expression) -> RType:
        dict_base_types = self.get_dict_base_type(expr)
        rtypes = [self.type_to_rtype(t.args[0]) for t in dict_base_types]
        return RUnion.make_simplified_union(rtypes)

    def get_dict_value_type(self, expr: Expression) -> RType:
        dict_base_types = self.get_dict_base_type(expr)
        rtypes = [self.type_to_rtype(t.args[1]) for t in dict_base_types]
        return RUnion.make_simplified_union(rtypes)

    def get_dict_item_type(self, expr: Expression) -> RType:
        key_type = self.get_dict_key_type(expr)
        value_type = self.get_dict_value_type(expr)
        return RTuple([key_type, value_type])

    def _analyze_iterable_item_type(self, expr: Expression) -> Type:
        """Return the item type given by 'expr' in an iterable context."""
        # This logic is copied from mypy's TypeChecker.analyze_iterable_item_type.
        if expr not in self.types:
            # Mypy thinks this is unreachable.
            iterable: ProperType = AnyType(TypeOfAny.from_error)
        else:
            iterable = get_proper_type(self.types[expr])
        echk = self.graph[self.module_name].type_checker().expr_checker
        iterator = echk.check_method_call_by_name("__iter__", iterable, [], [], expr)[0]

        from mypy.join import join_types

        if isinstance(iterable, TupleType):
            joined: Type = UninhabitedType()
            for item in iterable.items:
                joined = join_types(joined, item)
            return joined
        else:
            # Non-tuple iterable.
            return echk.check_method_call_by_name("__next__", iterator, [], [], expr)[0]

    def is_native_module(self, module: str) -> bool:
        """Is the given module one compiled by mypyc?"""
        return self.mapper.is_native_module(module)

    def is_native_ref_expr(self, expr: RefExpr) -> bool:
        return self.mapper.is_native_ref_expr(expr)

    def is_native_module_ref_expr(self, expr: RefExpr) -> bool:
        return self.mapper.is_native_module_ref_expr(expr)

    def is_synthetic_type(self, typ: TypeInfo) -> bool:
        """Is a type something other than just a class we've created?"""
        return typ.is_named_tuple or typ.is_newtype or typ.typeddict_type is not None

    def get_final_ref(self, expr: MemberExpr) -> tuple[str, Var, bool] | None:
        """Check if `expr` is a final attribute.

        This needs to be done differently for class and module attributes to
        correctly determine fully qualified name. Return a tuple that consists of
        the qualified name, the corresponding Var node, and a flag indicating whether
        the final name was defined in a compiled module. Return None if `expr` does not
        refer to a final attribute.
        """
        final_var = None
        if isinstance(expr.expr, RefExpr) and isinstance(expr.expr.node, TypeInfo):
            # a class attribute
            sym = expr.expr.node.get(expr.name)
            if sym and isinstance(sym.node, Var):
                # Enum attribute are treated as final since they are added to the global cache
                expr_fullname = expr.expr.node.bases[0].type.fullname
                is_final = sym.node.is_final or expr_fullname == "enum.Enum"
                if is_final:
                    final_var = sym.node
                    fullname = f"{sym.node.info.fullname}.{final_var.name}"
                    native = self.is_native_module(expr.expr.node.module_name)
        elif self.is_module_member_expr(expr):
            # a module attribute
            if isinstance(expr.node, Var) and expr.node.is_final:
                final_var = expr.node
                fullname = expr.node.fullname
                native = self.is_native_ref_expr(expr)
        if final_var is not None:
            return fullname, final_var, native
        return None

    def emit_load_final(
        self, final_var: Var, fullname: str, name: str, native: bool, typ: Type, line: int
    ) -> Value | None:
        """Emit code for loading value of a final name (if possible).

        Args:
            final_var: Var corresponding to the final name
            fullname: its qualified name
            name: shorter name to show in errors
            native: whether the name was defined in a compiled module
            typ: its type
            line: line number where loading occurs
        """
        if final_var.final_value is not None:  # this is safe even for non-native names
            return self.load_literal_value(final_var.final_value)
        elif native and module_prefix(self.graph, fullname):
            return self.load_final_static(fullname, self.mapper.type_to_rtype(typ), line, name)
        else:
            return None

    def is_module_member_expr(self, expr: MemberExpr) -> bool:
        return isinstance(expr.expr, RefExpr) and isinstance(expr.expr.node, MypyFile)

    def call_refexpr_with_args(
        self, expr: CallExpr, callee: RefExpr, arg_values: list[Value]
    ) -> Value:
        # Handle data-driven special-cased primitive call ops.
        if callee.fullname and expr.arg_kinds == [ARG_POS] * len(arg_values):
            fullname = get_call_target_fullname(callee)
            primitive_candidates = function_ops.get(fullname, [])
            target = self.builder.matching_primitive_op(
                primitive_candidates, arg_values, expr.line, self.node_type(expr)
            )
            if target:
                return target

        # Standard native call if signature and fullname are good and all arguments are positional
        # or named.
        callee_node = callee.node
        if isinstance(callee_node, OverloadedFuncDef):
            callee_node = callee_node.impl
        # TODO: use native calls for any decorated functions which have all their decorators
        # removed, not just singledispatch functions (which we don't do now just in case those
        # decorated functions are callable classes or cannot be called without the python API for
        # some other reason)
        if (
            isinstance(callee_node, Decorator)
            and callee_node.func not in self.fdefs_to_decorators
            and callee_node.func in self.singledispatch_impls
        ):
            callee_node = callee_node.func
        if (
            callee_node is not None
            and callee.fullname
            and callee_node in self.mapper.func_to_decl
            and all(kind in (ARG_POS, ARG_NAMED) for kind in expr.arg_kinds)
        ):
            decl = self.mapper.func_to_decl[callee_node]
            return self.builder.call(decl, arg_values, expr.arg_kinds, expr.arg_names, expr.line)

        # Fall back to a Python call
        function = self.accept(callee)
        return self.py_call(
            function, arg_values, expr.line, arg_kinds=expr.arg_kinds, arg_names=expr.arg_names
        )

    def shortcircuit_expr(self, expr: OpExpr) -> Value:
        return self.builder.shortcircuit_helper(
            expr.op,
            self.node_type(expr),
            lambda: self.accept(expr.left),
            lambda: self.accept(expr.right),
            expr.line,
        )

    # Basic helpers

    def flatten_classes(self, arg: RefExpr | TupleExpr) -> list[ClassIR] | None:
        """Flatten classes in isinstance(obj, (A, (B, C))).

        If at least one item is not a reference to a native class, return None.
        """
        if isinstance(arg, RefExpr):
            if isinstance(arg.node, TypeInfo) and self.is_native_module_ref_expr(arg):
                ir = self.mapper.type_to_ir.get(arg.node)
                if ir:
                    return [ir]
            return None
        else:
            res: list[ClassIR] = []
            for item in arg.items:
                if isinstance(item, (RefExpr, TupleExpr)):
                    item_part = self.flatten_classes(item)
                    if item_part is None:
                        return None
                    res.extend(item_part)
                else:
                    return None
            return res

    def enter(self, fn_info: FuncInfo | str = "", *, ret_type: RType = none_rprimitive) -> None:
        if isinstance(fn_info, str):
            fn_info = FuncInfo(name=fn_info)
        self.builder = LowLevelIRBuilder(self.errors, self.options)
        self.builder.set_module(self.module_name, self.module_path)
        self.builders.append(self.builder)
        self.symtables.append({})
        self.runtime_args.append([])
        self.fn_info = fn_info
        self.fn_infos.append(self.fn_info)
        self.ret_types.append(ret_type)
        if fn_info.is_generator:
            self.nonlocal_control.append(GeneratorNonlocalControl())
        else:
            self.nonlocal_control.append(BaseNonlocalControl())
        self.activate_block(BasicBlock())

    def leave(self) -> tuple[list[Register], list[RuntimeArg], list[BasicBlock], RType, FuncInfo]:
        builder = self.builders.pop()
        self.symtables.pop()
        runtime_args = self.runtime_args.pop()
        ret_type = self.ret_types.pop()
        fn_info = self.fn_infos.pop()
        self.nonlocal_control.pop()
        self.builder = self.builders[-1]
        self.fn_info = self.fn_infos[-1]
        return builder.args, runtime_args, builder.blocks, ret_type, fn_info

    @contextmanager
    def enter_method(
        self,
        class_ir: ClassIR,
        name: str,
        ret_type: RType,
        fn_info: FuncInfo | str = "",
        self_type: RType | None = None,
    ) -> Iterator[None]:
        """Generate IR for a method.

        If the method takes arguments, you should immediately afterwards call
        add_argument() for each non-self argument (self is created implicitly).

        Args:
            class_ir: Add method to this class
            name: Short name of the method
            ret_type: Return type of the method
            fn_info: Optionally, additional information about the method
            self_type: If not None, override default type of the implicit 'self'
                argument (by default, derive type from class_ir)
        """
        self.enter(fn_info, ret_type=ret_type)
        self.function_name_stack.append(name)
        self.class_ir_stack.append(class_ir)
        if self_type is None:
            self_type = RInstance(class_ir)
        self.add_argument(SELF_NAME, self_type)
        try:
            yield
        finally:
            arg_regs, args, blocks, ret_type, fn_info = self.leave()
            sig = FuncSignature(args, ret_type)
            name = self.function_name_stack.pop()
            class_ir = self.class_ir_stack.pop()
            decl = FuncDecl(name, class_ir.name, self.module_name, sig)
            ir = FuncIR(decl, arg_regs, blocks)
            class_ir.methods[name] = ir
            class_ir.method_decls[name] = ir.decl
            self.functions.append(ir)

    def add_argument(self, var: str | Var, typ: RType, kind: ArgKind = ARG_POS) -> Register:
        """Declare an argument in the current function.

        You should use this instead of directly calling add_local() in new code.
        """
        if isinstance(var, str):
            var = Var(var)
        reg = self.add_local(var, typ, is_arg=True)
        self.runtime_args[-1].append(RuntimeArg(var.name, typ, kind))
        return reg

    def lookup(self, symbol: SymbolNode) -> SymbolTarget:
        return self.symtables[-1][symbol]

    def add_local(self, symbol: SymbolNode, typ: RType, is_arg: bool = False) -> Register:
        """Add register that represents a symbol to the symbol table.

        Args:
            is_arg: is this a function argument
        """
        assert isinstance(symbol, SymbolNode)
        reg = Register(
            typ, remangle_redefinition_name(symbol.name), is_arg=is_arg, line=symbol.line
        )
        self.symtables[-1][symbol] = AssignmentTargetRegister(reg)
        if is_arg:
            self.builder.args.append(reg)
        return reg

    def add_local_reg(
        self, symbol: SymbolNode, typ: RType, is_arg: bool = False
    ) -> AssignmentTargetRegister:
        """Like add_local, but return an assignment target instead of value."""
        self.add_local(symbol, typ, is_arg)
        target = self.symtables[-1][symbol]
        assert isinstance(target, AssignmentTargetRegister)
        return target

    def add_self_to_env(self, cls: ClassIR) -> AssignmentTargetRegister:
        """Low-level function that adds a 'self' argument.

        This is only useful if using enter() instead of enter_method().
        """
        return self.add_local_reg(Var(SELF_NAME), RInstance(cls), is_arg=True)

    def add_target(self, symbol: SymbolNode, target: SymbolTarget) -> SymbolTarget:
        self.symtables[-1][symbol] = target
        return target

    def type_to_rtype(self, typ: Type | None) -> RType:
        return self.mapper.type_to_rtype(typ)

    def node_type(self, node: Expression) -> RType:
        if isinstance(node, IntExpr):
            # TODO: Don't special case IntExpr
            return int_rprimitive
        if node not in self.types:
            return object_rprimitive
        mypy_type = self.types[node]
        return self.type_to_rtype(mypy_type)

    def add_var_to_env_class(
        self,
        var: SymbolNode,
        rtype: RType,
        base: FuncInfo | ImplicitClass,
        reassign: bool = False,
        always_defined: bool = False,
    ) -> AssignmentTarget:
        # First, define the variable name as an attribute of the environment class, and then
        # construct a target for that attribute.
        name = remangle_redefinition_name(var.name)
        self.fn_info.env_class.attributes[name] = rtype
        if always_defined:
            self.fn_info.env_class.attrs_with_defaults.add(name)
        attr_target = AssignmentTargetAttr(base.curr_env_reg, name)

        if reassign:
            # Read the local definition of the variable, and set the corresponding attribute of
            # the environment class' variable to be that value.
            reg = self.read(self.lookup(var), self.fn_info.fitem.line)
            self.add(SetAttr(base.curr_env_reg, name, reg, self.fn_info.fitem.line))

        # Override the local definition of the variable to instead point at the variable in
        # the environment class.
        return self.add_target(var, attr_target)

    def is_builtin_ref_expr(self, expr: RefExpr) -> bool:
        assert expr.node, "RefExpr not resolved"
        return "." in expr.node.fullname and expr.node.fullname.split(".")[0] == "builtins"

    def load_global(self, expr: NameExpr) -> Value:
        """Loads a Python-level global.

        This takes a NameExpr and uses its name as a key to retrieve the corresponding PyObject *
        from the _globals dictionary in the C-generated code.
        """
        # If the global is from 'builtins', turn it into a module attr load instead
        if self.is_builtin_ref_expr(expr):
            assert expr.node, "RefExpr not resolved"
            return self.load_module_attr_by_fullname(expr.node.fullname, expr.line)
        if (
            self.is_native_module_ref_expr(expr)
            and isinstance(expr.node, TypeInfo)
            and not self.is_synthetic_type(expr.node)
        ):
            assert expr.fullname
            return self.load_native_type_object(expr.fullname)
        return self.load_global_str(expr.name, expr.line)

    def load_global_str(self, name: str, line: int) -> Value:
        _globals = self.load_globals_dict()
        reg = self.load_str(name)
        return self.primitive_op(dict_get_item_op, [_globals, reg], line)

    def load_globals_dict(self) -> Value:
        return self.add(LoadStatic(dict_rprimitive, "globals", self.module_name))

    def load_module_attr_by_fullname(self, fullname: str, line: int) -> Value:
        module, _, name = fullname.rpartition(".")
        left = self.load_module(module)
        return self.py_get_attr(left, name, line)

    def is_native_attr_ref(self, expr: MemberExpr) -> bool:
        """Is expr a direct reference to a native (struct) attribute of an instance?"""
        obj_rtype = self.node_type(expr.expr)
        return (
            isinstance(obj_rtype, RInstance)
            and obj_rtype.class_ir.is_ext_class
            and obj_rtype.class_ir.has_attr(expr.name)
            and not obj_rtype.class_ir.get_method(expr.name)
        )

    def mark_block_unreachable(self) -> None:
        """Mark statements in the innermost block being processed as unreachable.

        This should be called after a statement that unconditionally leaves the
        block, such as 'break' or 'return'.
        """
        self.block_reachable_stack[-1] = False

    # Lacks a good type because there wasn't a reasonable type in 3.5 :(
    def catch_errors(self, line: int) -> Any:
        return catch_errors(self.module_path, line)

    def warning(self, msg: str, line: int) -> None:
        self.errors.warning(msg, self.module_path, line)

    def error(self, msg: str, line: int) -> None:
        self.errors.error(msg, self.module_path, line)

    def note(self, msg: str, line: int) -> None:
        self.errors.note(msg, self.module_path, line)

    def add_function(self, func_ir: FuncIR, line: int) -> None:
        name = (func_ir.class_name, func_ir.name)
        if name in self.function_names:
            self.error(f'Duplicate definition of "{name[1]}" not supported by mypyc', line)
            return
        self.function_names.add(name)
        self.functions.append(func_ir)


def gen_arg_defaults(builder: IRBuilder) -> None:
    """Generate blocks for arguments that have default values.

    If the passed value is an error value, then assign the default
    value to the argument.
    """
    fitem = builder.fn_info.fitem
    nb = 0
    for arg in fitem.arguments:
        if arg.initializer:
            target = builder.lookup(arg.variable)

            def get_default() -> Value:
                assert arg.initializer is not None

                # If it is constant, don't bother storing it
                if is_constant(arg.initializer):
                    return builder.accept(arg.initializer)

                # Because gen_arg_defaults runs before calculate_arg_defaults, we
                # add the static/attribute to final_names/the class here.
                elif not builder.fn_info.is_nested:
                    name = fitem.fullname + "." + arg.variable.name
                    builder.final_names.append((name, target.type))
                    return builder.add(LoadStatic(target.type, name, builder.module_name))
                else:
                    name = arg.variable.name
                    builder.fn_info.callable_class.ir.attributes[name] = target.type
                    return builder.add(
                        GetAttr(builder.fn_info.callable_class.self_reg, name, arg.line)
                    )

            assert isinstance(target, AssignmentTargetRegister)
            reg = target.register
            if not reg.type.error_overlap:
                builder.assign_if_null(target.register, get_default, arg.initializer.line)
            else:
                builder.assign_if_bitmap_unset(
                    target.register, get_default, nb, arg.initializer.line
                )
                nb += 1


def remangle_redefinition_name(name: str) -> str:
    """Remangle names produced by mypy when allow-redefinition is used and a name
    is used with multiple types within a single block.

    We only need to do this for locals, because the name is used as the name of the register;
    for globals, the name itself is stored in a register for the purpose of doing dict
    lookups.
    """
    return name.replace("'", "__redef__")


def get_call_target_fullname(ref: RefExpr) -> str:
    if isinstance(ref.node, TypeAlias):
        # Resolve simple type aliases. In calls they evaluate to the type they point to.
        target = get_proper_type(ref.node.target)
        if isinstance(target, Instance):
            return target.type.fullname
    return ref.fullname


def create_type_params(
    builder: IRBuilder, typing_mod: Value, type_args: list[TypeParam], line: int
) -> list[Value]:
    """Create objects representing various kinds of Python 3.12 type parameters.

    The "typing_mod" argument is the "_typing" module object. The type objects
    are looked up from it.

    The returned list has one item for each "type_args" item, in the same order.
    Each item is either a TypeVar, TypeVarTuple or ParamSpec instance.
    """
    tvs = []
    type_var_imported: Value | None = None
    for type_param in type_args:
        if type_param.kind == TYPE_VAR_KIND:
            if type_var_imported:
                # Reuse previously imported value as a minor optimization
                tvt = type_var_imported
            else:
                tvt = builder.py_get_attr(typing_mod, "TypeVar", line)
                type_var_imported = tvt
        elif type_param.kind == TYPE_VAR_TUPLE_KIND:
            tvt = builder.py_get_attr(typing_mod, "TypeVarTuple", line)
        else:
            assert type_param.kind == PARAM_SPEC_KIND
            tvt = builder.py_get_attr(typing_mod, "ParamSpec", line)
        if type_param.kind != TYPE_VAR_TUPLE_KIND:
            # To match runtime semantics, pass infer_variance=True
            tv = builder.py_call(
                tvt,
                [builder.load_str(type_param.name), builder.true()],
                line,
                arg_kinds=[ARG_POS, ARG_NAMED],
                arg_names=[None, "infer_variance"],
            )
        else:
            tv = builder.py_call(tvt, [builder.load_str(type_param.name)], line)
        builder.init_type_var(tv, type_param.name, line)
        tvs.append(tv)
    return tvs


def calculate_arg_defaults(
    builder: IRBuilder,
    fn_info: FuncInfo,
    func_reg: Value | None,
    symtable: dict[SymbolNode, SymbolTarget],
) -> None:
    """Calculate default argument values and store them.

    They are stored in statics for top level functions and in
    the function objects for nested functions (while constants are
    still stored computed on demand).
    """
    fitem = fn_info.fitem
    for arg in fitem.arguments:
        # Constant values don't get stored but just recomputed
        if arg.initializer and not is_constant(arg.initializer):
            value = builder.coerce(
                builder.accept(arg.initializer), symtable[arg.variable].type, arg.line
            )
            if not fn_info.is_nested:
                name = fitem.fullname + "." + arg.variable.name
                builder.add(InitStatic(value, name, builder.module_name))
            else:
                assert func_reg is not None
                builder.add(SetAttr(func_reg, arg.variable.name, value, arg.line))



"""
Cloud Browser PDF Görüntüleyici
PDF dosyalarını görüntüleme desteği
"""

import os
from pathlib import Path
from typing import Optional

try:
    from PyQt6.QtWidgets import *
    from PyQt6.QtCore import *
    from PyQt6.QtGui import *
    from PyQt6.QtWebEngineWidgets import *
    from PyQt6.QtWebEngineCore import *
    from PyQt6.QtPrintSupport import QPrinter, QPrintDialog
    WEBENGINE_AVAILABLE = True
except ImportError:
    try:
        from PyQt6.QtWidgets import *
        from PyQt6.QtCore import *
        from PyQt6.QtGui import *
        from PyQt6.QtPrintSupport import QPrinter, QPrintDialog
        WEBENGINE_AVAILABLE = False
    except ImportError:
        raise ImportError("PyQt6 is required for Cloud Browser")

class PDFViewer(QWidget):
    """
    PDF görüntüleyici widget'ı
    """
    
    def __init__(self, parent=None):
        super().__init__(parent)
        self.current_pdf_path = None
        self.current_page = 1
        self.total_pages = 1
        self.zoom_factor = 1.0
        
        self.init_ui()
    
    def init_ui(self):
        """UI'yı başlat"""
        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        
        # PDF toolbar
        self.create_pdf_toolbar()
        layout.addWidget(self.pdf_toolbar)
        
        # PDF görüntüleme alanı
        if WEBENGINE_AVAILABLE:
            # WebEngine ile PDF görüntüleme
            self.pdf_view = QWebEngineView()
            self.pdf_view.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
            self.pdf_view.customContextMenuRequested.connect(self.show_pdf_context_menu)
        else:
            # Basit metin görüntüleme
            self.pdf_view = QTextEdit()
            self.pdf_view.setReadOnly(True)
            self.pdf_view.setHtml("""
                <h2>PDF Görüntüleyici</h2>
                <p>WebEngine mevcut değil. PDF dosyaları görüntülenemiyor.</p>
                <p>PDF desteği için PyQt6-WebEngine gereklidir:</p>
                <code>pip install PyQt6-WebEngine</code>
            """)
        
        layout.addWidget(self.pdf_view)
        
        # Status bar
        self.status_label = QLabel("PDF yüklenmedi")
        layout.addWidget(self.status_label)
    
    def create_pdf_toolbar(self):
        """PDF toolbar oluştur"""
        self.pdf_toolbar = QToolBar()
        self.pdf_toolbar.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextBesideIcon)
        
        # Dosya aç
        open_action = QAction("📂 Aç", self)
        open_action.setToolTip("PDF dosyası aç")
        open_action.triggered.connect(self.open_pdf_file)
        self.pdf_toolbar.addAction(open_action)
        
        self.pdf_toolbar.addSeparator()
        
        # Sayfa navigasyonu
        self.prev_page_action = QAction("⬅️ Önceki", self)
        self.prev_page_action.setToolTip("Önceki sayfa")
        self.prev_page_action.triggered.connect(self.prev_page)
        self.prev_page_action.setEnabled(False)
        self.pdf_toolbar.addAction(self.prev_page_action)
        
        # Sayfa numarası
        self.page_label = QLabel("Sayfa: 0 / 0")
        self.pdf_toolbar.addWidget(self.page_label)
        
        self.next_page_action = QAction("➡️ Sonraki", self)
        self.next_page_action.setToolTip("Sonraki sayfa")
        self.next_page_action.triggered.connect(self.next_page)
        self.next_page_action.setEnabled(False)
        self.pdf_toolbar.addAction(self.next_page_action)
        
        self.pdf_toolbar.addSeparator()
        
        # Zoom kontrolleri
        zoom_out_action = QAction("🔍-", self)
        zoom_out_action.setToolTip("Uzaklaştır")
        zoom_out_action.triggered.connect(self.zoom_out)
        self.pdf_toolbar.addAction(zoom_out_action)
        
        self.zoom_label = QLabel("100%")
        self.zoom_label.setMinimumWidth(50)
        self.pdf_toolbar.addWidget(self.zoom_label)
        
        zoom_in_action = QAction("🔍+", self)
        zoom_in_action.setToolTip("Yakınlaştır")
        zoom_in_action.triggered.connect(self.zoom_in)
        self.pdf_toolbar.addAction(zoom_in_action)
        
        zoom_fit_action = QAction("📄 Sığdır", self)
        zoom_fit_action.setToolTip("Sayfaya sığdır")
        zoom_fit_action.triggered.connect(self.zoom_fit)
        self.pdf_toolbar.addAction(zoom_fit_action)
        
        self.pdf_toolbar.addSeparator()
        
        # Yazdır
        print_action = QAction("🖨️ Yazdır", self)
        print_action.setToolTip("PDF'i yazdır")
        print_action.triggered.connect(self.print_pdf)
        self.pdf_toolbar.addAction(print_action)
        
        # Stil uygula
        self.pdf_toolbar.setStyleSheet("""
            QToolBar {
                background-color: #f8f9fa;
                border: none;
                border-bottom: 1px solid #dee2e6;
                padding: 4px;
            }
            QToolBar QAction {
                padding: 6px 12px;
                margin: 2px;
                border-radius: 4px;
            }
            QToolBar QAction:hover {
                background-color: #e9ecef;
            }
            QLabel {
                padding: 4px 8px;
                color: #6c757d;
            }
        """)
    
    def load_pdf(self, file_path: str):
        """PDF dosyası yükle"""
        if not os.path.exists(file_path):
            self.show_error("Dosya bulunamadı", f"PDF dosyası bulunamadı: {file_path}")
            return False
        
        if not file_path.lower().endswith('.pdf'):
            self.show_error("Geçersiz dosya", "Bu dosya bir PDF değil.")
            return False
        
        try:
            self.current_pdf_path = file_path
            
            if WEBENGINE_AVAILABLE:
                # WebEngine ile PDF yükle
                file_url = QUrl.fromLocalFile(os.path.abspath(file_path))
                self.pdf_view.setUrl(file_url)
                
                # PDF yükleme tamamlandığında
                self.pdf_view.loadFinished.connect(self.pdf_loaded)
            else:
                # Basit metin gösterimi
                self.pdf_view.setHtml(f"""
                    <h2>PDF Dosyası</h2>
                    <p><b>Dosya:</b> {os.path.basename(file_path)}</p>
                    <p><b>Yol:</b> {file_path}</p>
                    <p><b>Boyut:</b> {self.format_file_size(os.path.getsize(file_path))}</p>
                    <br>
                    <p>WebEngine mevcut olmadığı için PDF içeriği görüntülenemiyor.</p>
                    <p>PDF desteği için PyQt6-WebEngine gereklidir.</p>
                """)
            
            # UI güncelle
            self.update_ui()
            return True
            
        except Exception as e:
            self.show_error("PDF Yükleme Hatası", f"PDF yüklenirken hata oluştu: {e}")
            return False
    
    def pdf_loaded(self, success):
        """PDF yükleme tamamlandığında"""
        if success:
            self.status_label.setText(f"PDF yüklendi: {os.path.basename(self.current_pdf_path)}")
            # TODO: Sayfa sayısını al (WebEngine API'si gerekli)
            self.total_pages = 1  # Placeholder
            self.current_page = 1
            self.update_page_info()
        else:
            self.show_error("PDF Yükleme Hatası", "PDF dosyası yüklenemedi.")
    
    def open_pdf_file(self):
        """PDF dosyası seç ve aç"""
        file_path, _ = QFileDialog.getOpenFileName(
            self,
            "PDF Dosyası Aç",
            "",
            "PDF Dosyaları (*.pdf);;Tüm Dosyalar (*.*)"
        )
        
        if file_path:
            self.load_pdf(file_path)
    
    def prev_page(self):
        """Önceki sayfa"""
        if self.current_page > 1:
            self.current_page -= 1
            self.update_page_info()
            # TODO: WebEngine'de sayfa değiştirme implementasyonu
    
    def next_page(self):
        """Sonraki sayfa"""
        if self.current_page < self.total_pages:
            self.current_page += 1
            self.update_page_info()
            # TODO: WebEngine'de sayfa değiştirme implementasyonu
    
    def zoom_in(self):
        """Yakınlaştır"""
        self.zoom_factor = min(self.zoom_factor * 1.2, 5.0)
        self.apply_zoom()
    
    def zoom_out(self):
        """Uzaklaştır"""
        self.zoom_factor = max(self.zoom_factor / 1.2, 0.2)
        self.apply_zoom()
    
    def zoom_fit(self):
        """Sayfaya sığdır"""
        self.zoom_factor = 1.0
        self.apply_zoom()
    
    def apply_zoom(self):
        """Zoom uygula"""
        if WEBENGINE_AVAILABLE and hasattr(self.pdf_view, 'setZoomFactor'):
            self.pdf_view.setZoomFactor(self.zoom_factor)
        
        # Zoom label güncelle
        zoom_percent = int(self.zoom_factor * 100)
        self.zoom_label.setText(f"{zoom_percent}%")
    
    def print_pdf(self):
        """PDF'i yazdır"""
        if not self.current_pdf_path:
            self.show_error("Hata", "Yazdırılacak PDF dosyası yok.")
            return
        
        if WEBENGINE_AVAILABLE:
            # WebEngine print dialog
            printer = QPrinter()
            print_dialog = QPrintDialog(printer, self)
            
            if print_dialog.exec() == QDialog.DialogCode.Accepted:
                # TODO: WebEngine print implementasyonu
                QMessageBox.information(self, "Yazdırma", "Yazdırma özelliği henüz implementasyonda...")
        else:
            QMessageBox.information(self, "Yazdırma", "Yazdırma için WebEngine gereklidir.")
    
    def show_pdf_context_menu(self, position):
        """PDF context menü"""
        if not WEBENGINE_AVAILABLE:
            return
        
        menu = QMenu(self)
        
        # Kopyala
        copy_action = menu.addAction("📋 Kopyala")
        copy_action.triggered.connect(self.copy_selection)
        
        # Seç
        select_all_action = menu.addAction("🔘 Tümünü Seç")
        select_all_action.triggered.connect(self.select_all)
        
        menu.addSeparator()
        
        # Zoom
        zoom_in_action = menu.addAction("🔍+ Yakınlaştır")
        zoom_in_action.triggered.connect(self.zoom_in)
        
        zoom_out_action = menu.addAction("🔍- Uzaklaştır")
        zoom_out_action.triggered.connect(self.zoom_out)
        
        zoom_fit_action = menu.addAction("📄 Sığdır")
        zoom_fit_action.triggered.connect(self.zoom_fit)
        
        menu.addSeparator()
        
        # Yazdır
        print_action = menu.addAction("🖨️ Yazdır")
        print_action.triggered.connect(self.print_pdf)
        
        menu.exec(self.pdf_view.mapToGlobal(position))
    
    def copy_selection(self):
        """Seçili metni kopyala"""
        if WEBENGINE_AVAILABLE and hasattr(self.pdf_view, 'page'):
            # TODO: WebEngine selection copy implementasyonu
            pass
    
    def select_all(self):
        """Tümünü seç"""
        if WEBENGINE_AVAILABLE and hasattr(self.pdf_view, 'page'):
            # TODO: WebEngine select all implementasyonu
            pass
    
    def update_ui(self):
        """UI'yı güncelle"""
        has_pdf = self.current_pdf_path is not None
        
        # Sayfa navigasyon butonları
        self.prev_page_action.setEnabled(has_pdf and self.current_page > 1)
        self.next_page_action.setEnabled(has_pdf and self.current_page < self.total_pages)
        
        # Sayfa bilgisi
        if has_pdf:
            self.update_page_info()
        else:
            self.page_label.setText("Sayfa: 0 / 0")
    
    def update_page_info(self):
        """Sayfa bilgisini güncelle"""
        self.page_label.setText(f"Sayfa: {self.current_page} / {self.total_pages}")
        
        # Navigasyon butonları
        self.prev_page_action.setEnabled(self.current_page > 1)
        self.next_page_action.setEnabled(self.current_page < self.total_pages)
    
    def show_error(self, title: str, message: str):
        """Hata mesajı göster"""
        QMessageBox.critical(self, title, message)
        self.status_label.setText(f"Hata: {message}")
    
    def format_file_size(self, size_bytes: int) -> str:
        """Dosya boyutunu formatla"""
        if size_bytes == 0:
            return "0 B"
        
        size_names = ["B", "KB", "MB", "GB"]
        i = 0
        while size_bytes >= 1024 and i < len(size_names) - 1:
            size_bytes /= 1024.0
            i += 1
        
        return f"{size_bytes:.1f} {size_names[i]}"

class PDFTab(QWidget):
    """
    PDF sekmesi için özel widget
    """
    
    def __init__(self, pdf_path: str, parent=None):
        super().__init__(parent)
        self.pdf_path = pdf_path
        
        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        
        # PDF viewer
        self.pdf_viewer = PDFViewer(self)
        layout.addWidget(self.pdf_viewer)
        
        # PDF'i yükle
        self.pdf_viewer.load_pdf(pdf_path)
    
    def get_title(self) -> str:
        """Sekme başlığını al"""
        if self.pdf_path:
            return f"📄 {os.path.basename(self.pdf_path)}"
        return "📄 PDF"
    
    def get_url(self) -> str:
        """URL'yi al"""
        if self.pdf_path:
            return f"file://{self.pdf_path}"
        return "about:blank"

def is_pdf_url(url: str) -> bool:
    """URL PDF dosyası mı kontrol et"""
    if not url:
        return False
    
    # URL'den dosya uzantısını al
    if url.lower().endswith('.pdf'):
        return True
    
    # MIME type kontrolü (gelecekte implementasyon)
    # TODO: HTTP response header'larından MIME type kontrolü
    
    return False

def create_pdf_tab(url: str, parent=None) -> Optional[PDFTab]:
    """PDF sekmesi oluştur"""
    if not is_pdf_url(url):
        return None
    
    # file:// URL'sini dosya yoluna çevir
    if url.startswith('file://'):
        file_path = url[7:]  # file:// kısmını kaldır
        if os.path.exists(file_path):
            return PDFTab(file_path, parent)
    
    # HTTP/HTTPS URL'leri için (gelecekte implementasyon)
    # TODO: PDF dosyasını indir ve geçici dosya olarak aç
    
    return None 


#!/usr/bin/env python3
"""
Cloud Notepad v2.0.0 - Modern Metin Düzenleyici
PyCloud OS için geliştirilmiş sekmeli metin düzenleyici
"""

import sys
import os
import logging
import argparse
from pathlib import Path

# PyQt6 import ve gerekli ayarlar
try:
    from PyQt6.QtWidgets import QApplication
    from PyQt6.QtCore import Qt, QCoreApplication
    from PyQt6.QtGui import QIcon
    PYQT_AVAILABLE = True
except ImportError as e:
    print(f"❌ PyQt6 import hatası: {e}")
    print("💡 Çözüm: pip install PyQt6")
    PYQT_AVAILABLE = False

# PyCloud OS core modüllerini ekle
project_root = Path(__file__).parent.parent.parent
sys.path.insert(0, str(project_root))

def setup_logging():
    """Logging sistemini kur"""
    log_format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    logging.basicConfig(
        level=logging.INFO,
        format=log_format,
        handlers=[
            logging.StreamHandler(sys.stdout)
        ]
    )
    return logging.getLogger("CloudNotepad")

def main():
    """Ana fonksiyon"""
    logger = setup_logging()
    logger.info("📝 Cloud Notepad v2.0.0 starting...")
    
    if not PYQT_AVAILABLE:
        logger.error("❌ PyQt6 not available")
        return 1
    
    # Komut satırı parametrelerini parse et
    parser = argparse.ArgumentParser(description='PyCloud OS Cloud Notepad v2.0.0')
    parser.add_argument('--open-file', type=str, help='Açılacak dosya yolu')
    parser.add_argument('--open-path', type=str, help='Açılacak dizin yolu (dosya için)')
    args, unknown = parser.parse_known_args()
    
    # QApplication oluştur veya var olanı kullan
    app = QApplication.instance()
    if app is None:
        app = QApplication(sys.argv)
        app.setApplicationName("Cloud Notepad")
        app.setApplicationVersion("2.0.0")
        app.setOrganizationName("PyCloud OS")
        
        # High DPI desteği
        try:
            app.setAttribute(Qt.ApplicationAttribute.AA_EnableHighDpiScaling, True)
            app.setAttribute(Qt.ApplicationAttribute.AA_UseHighDpiPixmaps, True)
        except AttributeError:
            pass  # PyQt6'da varsayılan olarak aktif
    
    # Kernel referansı - Bridge sistemi üzerinden al
    kernel = None
    try:
        # Bridge sistemi üzerinden kernel'e erişim
        from core.bridge import get_bridge_manager
        bridge_manager = get_bridge_manager()
        if bridge_manager:
            kernel = bridge_manager.get_kernel_reference()
            print("✅ Kernel referansı bridge üzerinden alındı")
            
            # Bridge bağlantısını test et
            if hasattr(bridge_manager, 'call_module_method'):
                # IPC client kullanılıyor
                print("🔗 Bridge IPC client kullanılıyor")
            else:
                # Direkt bridge manager kullanılıyor
                print("🔗 Bridge manager direkt kullanılıyor")
                
        else:
            print("⚠️ Bridge manager bulunamadı, standalone modda çalışılıyor")
    except Exception as e:
        print(f"⚠️ Bridge bağlantısı kurulamadı: {e}")
        print("Standalone modda çalışılıyor...")
    
    try:
        # Notepad'ı oluştur
        logger.info("Creating Cloud Notepad instance...")
        from cloud.notepad import create_notepad_app
        
        notepad = create_notepad_app(kernel)
        
        # Icon ayarla (varsa)
        icon_path = Path(__file__).parent / "icon.png"
        if icon_path.exists():
            notepad.setWindowIcon(QIcon(str(icon_path)))
        
        # Dosya açma parametresi varsa dosyayı aç
        if args.open_file:
            # PyCloud OS sanal dosya sistemi yollarını destekle
            file_path_str = args.open_file
            
            # Gerçek dosya sistemi yolu mu kontrol et
            file_path = Path(file_path_str)
            file_exists = file_path.exists()
            
            # PyCloud OS sanal dosya sistemi yolu da olabilir
            if not file_exists and not file_path_str.startswith('/'):
                # Sanal dosya sistemi yolu olarak kabul et
                logger.info(f"PyCloud OS sanal dosya sistemi yolu olarak açılıyor: {file_path_str}")
                notepad.open_specific_file(file_path_str)
                logger.info(f"Opened file: {args.open_file}")
            elif file_exists:
                # Gerçek dosya sistemi yolu
                logger.info(f"Gerçek dosya sistemi yolu olarak açılıyor: {file_path_str}")
                notepad.open_specific_file(str(file_path))
                logger.info(f"Opened file: {args.open_file}")
            else:
                logger.warning(f"File not found: {args.open_file}")
        
        logger.info("Showing notepad window...")
        notepad.show()
        
        logger.info("✅ Cloud Notepad v2.0.0 started successfully")
        logger.info("Starting event loop...")
        
        # Event loop başlat
        exit_code = app.exec()
        
        logger.info(f"🔚 Cloud Notepad exited with code: {exit_code}")
        return exit_code
        
    except Exception as e:
        logger.error(f"❌ Critical error: {e}")
        import traceback
        traceback.print_exc()
        return 1
    
    finally:
        logger.info("🧹 Cleaning up...")

if __name__ == "__main__":
    sys.exit(main()) 


import sys
from optparse import Values
from pathlib import Path
from typing import List

from pip._internal.cache import WheelCache
from pip._internal.cli import cmdoptions
from pip._internal.cli.req_command import (
    RequirementCommand,
    with_cleanup,
)
from pip._internal.cli.status_codes import SUCCESS
from pip._internal.models.pylock import Pylock, is_valid_pylock_file_name
from pip._internal.operations.build.build_tracker import get_build_tracker
from pip._internal.req.req_install import (
    check_legacy_setup_py_options,
)
from pip._internal.utils.logging import getLogger
from pip._internal.utils.misc import (
    get_pip_version,
)
from pip._internal.utils.temp_dir import TempDirectory

logger = getLogger(__name__)


class LockCommand(RequirementCommand):
    """
    EXPERIMENTAL - Lock packages and their dependencies from:

    - PyPI (and other indexes) using requirement specifiers.
    - VCS project urls.
    - Local project directories.
    - Local or remote source archives.

    pip also supports locking from "requirements files", which provide an easy
    way to specify a whole environment to be installed.

    The generated lock file is only guaranteed to be valid for the current
    python version and platform.
    """

    usage = """
      %prog [options] [-e] <local project path> ...
      %prog [options] <requirement specifier> [package-index-options] ...
      %prog [options] -r <requirements file> [package-index-options] ...
      %prog [options] <archive url/path> ..."""

    def add_options(self) -> None:
        self.cmd_opts.add_option(
            cmdoptions.PipOption(
                "--output",
                "-o",
                dest="output_file",
                metavar="path",
                type="path",
                default="pylock.toml",
                help="Lock file name (default=pylock.toml). Use - for stdout.",
            )
        )
        self.cmd_opts.add_option(cmdoptions.requirements())
        self.cmd_opts.add_option(cmdoptions.constraints())
        self.cmd_opts.add_option(cmdoptions.no_deps())
        self.cmd_opts.add_option(cmdoptions.pre())

        self.cmd_opts.add_option(cmdoptions.editable())

        self.cmd_opts.add_option(cmdoptions.src())

        self.cmd_opts.add_option(cmdoptions.ignore_requires_python())
        self.cmd_opts.add_option(cmdoptions.no_build_isolation())
        self.cmd_opts.add_option(cmdoptions.use_pep517())
        self.cmd_opts.add_option(cmdoptions.no_use_pep517())
        self.cmd_opts.add_option(cmdoptions.check_build_deps())

        self.cmd_opts.add_option(cmdoptions.config_settings())

        self.cmd_opts.add_option(cmdoptions.no_binary())
        self.cmd_opts.add_option(cmdoptions.only_binary())
        self.cmd_opts.add_option(cmdoptions.prefer_binary())
        self.cmd_opts.add_option(cmdoptions.require_hashes())
        self.cmd_opts.add_option(cmdoptions.progress_bar())

        index_opts = cmdoptions.make_option_group(
            cmdoptions.index_group,
            self.parser,
        )

        self.parser.insert_option_group(0, index_opts)
        self.parser.insert_option_group(0, self.cmd_opts)

    @with_cleanup
    def run(self, options: Values, args: List[str]) -> int:
        logger.verbose("Using %s", get_pip_version())

        logger.warning(
            "pip lock is currently an experimental command. "
            "It may be removed/changed in a future release "
            "without prior warning."
        )

        session = self.get_default_session(options)

        finder = self._build_package_finder(
            options=options,
            session=session,
            ignore_requires_python=options.ignore_requires_python,
        )
        build_tracker = self.enter_context(get_build_tracker())

        directory = TempDirectory(
            delete=not options.no_clean,
            kind="install",
            globally_managed=True,
        )

        reqs = self.get_requirements(args, options, finder, session)
        check_legacy_setup_py_options(options, reqs)

        wheel_cache = WheelCache(options.cache_dir)

        # Only when installing is it permitted to use PEP 660.
        # In other circumstances (pip wheel, pip download) we generate
        # regular (i.e. non editable) metadata and wheels.
        for req in reqs:
            req.permit_editable_wheels = True

        preparer = self.make_requirement_preparer(
            temp_build_dir=directory,
            options=options,
            build_tracker=build_tracker,
            session=session,
            finder=finder,
            use_user_site=False,
            verbosity=self.verbosity,
        )
        resolver = self.make_resolver(
            preparer=preparer,
            finder=finder,
            options=options,
            wheel_cache=wheel_cache,
            use_user_site=False,
            ignore_installed=True,
            ignore_requires_python=options.ignore_requires_python,
            upgrade_strategy="to-satisfy-only",
            use_pep517=options.use_pep517,
        )

        self.trace_basic_info(finder)

        requirement_set = resolver.resolve(reqs, check_supported_wheels=True)

        if options.output_file == "-":
            base_dir = Path.cwd()
        else:
            output_file_path = Path(options.output_file)
            if not is_valid_pylock_file_name(output_file_path):
                logger.warning(
                    "%s is not a valid lock file name.",
                    output_file_path,
                )
            base_dir = output_file_path.parent
        pylock_toml = Pylock.from_install_requirements(
            requirement_set.requirements.values(), base_dir=base_dir
        ).as_toml()
        if options.output_file == "-":
            sys.stdout.write(pylock_toml)
        else:
            output_file_path.write_text(pylock_toml, encoding="utf-8")

        return SUCCESS



"""
PyCloud OS Cloud Settings
Rain arayüzü için özelleştirme ve sistemsel ayarları yöneten uygulama
"""

import os
import sys
import json
import logging
from pathlib import Path
from typing import Dict, List, Optional, Any
from datetime import datetime

try:
    from PyQt6.QtWidgets import *
    from PyQt6.QtCore import *
    from PyQt6.QtGui import *
except ImportError:
    print("PyQt6 not available for Cloud Settings")
    sys.exit(1)

class SettingsPage(QWidget):
    """Ayar sayfası temel sınıfı"""
    
    settings_changed = pyqtSignal(str, dict)  # category, settings
    
    def __init__(self, title: str, icon: str = "⚙️"):
        super().__init__()
        self.title = title
        self.icon = icon
        self.settings = {}
        
        self.setup_ui()
    
    def setup_ui(self):
        """UI kurulumu - alt sınıflarda override edilecek"""
        layout = QVBoxLayout(self)
        
        title_label = QLabel(f"{self.icon} {self.title}")
        title_label.setStyleSheet("font-size: 18px; font-weight: bold; margin-bottom: 20px;")
        layout.addWidget(title_label)
        
        layout.addStretch()
    
    def load_settings(self, settings: Dict):
        """Ayarları yükle"""
        self.settings = settings
        self.update_ui()
    
    def update_ui(self):
        """UI'yi ayarlara göre güncelle"""
        pass
    
    def get_settings(self) -> Dict:
        """Mevcut ayarları al"""
        return self.settings.copy()

class AppearancePage(SettingsPage):
    """Görünüm ayarları sayfası"""
    
    def __init__(self, kernel=None):
        self.kernel = kernel
        self.logger = logging.getLogger("AppearancePage")
        super().__init__("Görünüm", "🎨")
    
    def setup_ui(self):
        """UI kurulumu"""
        layout = QVBoxLayout(self)
        
        # Başlık
        title_label = QLabel(f"{self.icon} {self.title}")
        title_label.setStyleSheet("font-size: 18px; font-weight: bold; margin-bottom: 20px;")
        layout.addWidget(title_label)
        
        # Tema seçimi
        theme_group = QGroupBox("Tema")
        theme_layout = QVBoxLayout(theme_group)
        
        self.theme_light = QRadioButton("Açık Tema")
        self.theme_dark = QRadioButton("Koyu Tema")
        self.theme_auto = QRadioButton("Sistem Teması")
        
        theme_layout.addWidget(self.theme_light)
        theme_layout.addWidget(self.theme_dark)
        theme_layout.addWidget(self.theme_auto)
        
        layout.addWidget(theme_group)
        
        # Duvar kağıdı
        wallpaper_group = QGroupBox("Duvar Kağıdı")
        wallpaper_layout = QVBoxLayout(wallpaper_group)
        
        wallpaper_select_layout = QHBoxLayout()
        self.wallpaper_path = QLineEdit()
        self.wallpaper_path.setPlaceholderText("Duvar kağıdı dosyası seçin...")
        wallpaper_browse = QPushButton("Gözat")
        wallpaper_browse.clicked.connect(self.browse_wallpaper)
        
        wallpaper_select_layout.addWidget(self.wallpaper_path)
        wallpaper_select_layout.addWidget(wallpaper_browse)
        wallpaper_layout.addLayout(wallpaper_select_layout)
        
        # Duvar kağıdı modu
        wallpaper_mode_layout = QHBoxLayout()
        wallpaper_mode_layout.addWidget(QLabel("Mod:"))
        
        self.wallpaper_mode = QComboBox()
        self.wallpaper_mode.addItems(["Sığdır", "Uzat", "Döşe", "Merkez", "Doldur"])
        wallpaper_mode_layout.addWidget(self.wallpaper_mode)
        wallpaper_mode_layout.addStretch()
        
        wallpaper_layout.addLayout(wallpaper_mode_layout)
        layout.addWidget(wallpaper_group)
        
        # Dock ayarları
        dock_group = QGroupBox("Dock")
        dock_layout = QFormLayout(dock_group)
        
        self.dock_position = QComboBox()
        self.dock_position.addItems(["Alt", "Sol", "Sağ", "Üst"])
        dock_layout.addRow("Konum:", self.dock_position)
        
        self.dock_size = QSlider(Qt.Orientation.Horizontal)
        self.dock_size.setRange(32, 128)
        self.dock_size.setValue(64)
        self.dock_size_label = QLabel("64px")
        self.dock_size.valueChanged.connect(lambda v: self.dock_size_label.setText(f"{v}px"))
        
        dock_size_layout = QHBoxLayout()
        dock_size_layout.addWidget(self.dock_size)
        dock_size_layout.addWidget(self.dock_size_label)
        dock_layout.addRow("Boyut:", dock_size_layout)
        
        self.dock_autohide = QCheckBox("Otomatik gizle")
        dock_layout.addRow("", self.dock_autohide)
        
        layout.addWidget(dock_group)
        
        # Topbar ayarları
        topbar_group = QGroupBox("Üst Çubuk")
        topbar_layout = QFormLayout(topbar_group)
        
        self.show_clock = QCheckBox("Saati göster")
        self.show_clock.setChecked(True)
        topbar_layout.addRow("", self.show_clock)
        
        self.show_user = QCheckBox("Kullanıcı menüsünü göster")
        self.show_user.setChecked(True)
        topbar_layout.addRow("", self.show_user)
        
        self.show_notifications = QCheckBox("Bildirimleri göster")
        self.show_notifications.setChecked(True)
        topbar_layout.addRow("", self.show_notifications)
        
        layout.addWidget(topbar_group)
        
        layout.addStretch()
        
        # Sinyal bağlantıları
        self.theme_light.toggled.connect(self.on_settings_changed)
        self.theme_dark.toggled.connect(self.on_settings_changed)
        self.theme_auto.toggled.connect(self.on_settings_changed)
        self.wallpaper_path.textChanged.connect(self.on_settings_changed)
        self.wallpaper_mode.currentTextChanged.connect(self.on_settings_changed)
        self.dock_position.currentTextChanged.connect(self.on_settings_changed)
        self.dock_size.valueChanged.connect(self.on_settings_changed)
        self.dock_autohide.toggled.connect(self.on_settings_changed)
        self.show_clock.toggled.connect(self.on_settings_changed)
        self.show_user.toggled.connect(self.on_settings_changed)
        self.show_notifications.toggled.connect(self.on_settings_changed)
    
    def browse_wallpaper(self):
        """Duvar kağıdı dosyası seç"""
        file_path, _ = QFileDialog.getOpenFileName(
            self, "Duvar Kağıdı Seç",
            str(Path.home()),
            "Resim Dosyaları (*.png *.jpg *.jpeg *.bmp *.gif *.webp)"
        )
        
        if file_path:
            # Wallpaper manager'ı kullanarak ayarla
            try:
                if self.kernel:
                    wallpaper_manager = self.kernel.get_module("wallpaper")
                    if wallpaper_manager:
                        # Önce system/wallpapers'a kopyala
                        if wallpaper_manager.add_wallpaper(file_path):
                            # Sonra yeni dosyayı ayarla
                            new_file_name = Path(file_path).name
                            system_wallpaper_path = Path("system/wallpapers") / new_file_name
                            
                            if wallpaper_manager.set_wallpaper(str(system_wallpaper_path)):
                                self.wallpaper_path.setText(str(system_wallpaper_path))
                                QMessageBox.information(self, "Başarılı", "Duvar kağıdı başarıyla ayarlandı!")
                            else:
                                self.wallpaper_path.setText(file_path)
                                QMessageBox.warning(self, "Uyarı", "Duvar kağıdı ayarlanamadı, sadece yol güncellendi.")
                        else:
                            # Direkt dosya yolunu kullan
                            self.wallpaper_path.setText(file_path)
                    else:
                        # Wallpaper manager yoksa sadece yol ayarla
                        self.wallpaper_path.setText(file_path)
                else:
                    self.wallpaper_path.setText(file_path)
                    
            except Exception as e:
                self.logger.error(f"Wallpaper setting error: {e}")
                self.wallpaper_path.setText(file_path)
    
    def on_settings_changed(self):
        """Ayar değişti"""
        settings = {
            "theme": "light" if self.theme_light.isChecked() else "dark" if self.theme_dark.isChecked() else "auto",
            "wallpaper_path": self.wallpaper_path.text(),
            "wallpaper_mode": self.wallpaper_mode.currentText().lower(),
            "dock_position": self.dock_position.currentText().lower(),
            "dock_size": self.dock_size.value(),
            "dock_autohide": self.dock_autohide.isChecked(),
            "show_clock": self.show_clock.isChecked(),
            "show_user": self.show_user.isChecked(),
            "show_notifications": self.show_notifications.isChecked()
        }
        
        self.settings = settings
        self.settings_changed.emit("appearance", settings)
    
    def update_ui(self):
        """UI'yi ayarlara göre güncelle"""
        theme = self.settings.get("theme", "auto")
        if theme == "light":
            self.theme_light.setChecked(True)
        elif theme == "dark":
            self.theme_dark.setChecked(True)
        else:
            self.theme_auto.setChecked(True)
        
        self.wallpaper_path.setText(self.settings.get("wallpaper_path", ""))
        
        wallpaper_mode = self.settings.get("wallpaper_mode", "sığdır")
        index = self.wallpaper_mode.findText(wallpaper_mode.title())
        if index >= 0:
            self.wallpaper_mode.setCurrentIndex(index)
        
        dock_position = self.settings.get("dock_position", "alt")
        index = self.dock_position.findText(dock_position.title())
        if index >= 0:
            self.dock_position.setCurrentIndex(index)
        
        self.dock_size.setValue(self.settings.get("dock_size", 64))
        self.dock_autohide.setChecked(self.settings.get("dock_autohide", False))
        self.show_clock.setChecked(self.settings.get("show_clock", True))
        self.show_user.setChecked(self.settings.get("show_user", True))
        self.show_notifications.setChecked(self.settings.get("show_notifications", True))

class SystemPage(SettingsPage):
    """Sistem ayarları sayfası"""
    
    def __init__(self):
        super().__init__("Sistem", "💻")
    
    def setup_ui(self):
        """UI kurulumu"""
        layout = QVBoxLayout(self)
        
        # Başlık
        title_label = QLabel(f"{self.icon} {self.title}")
        title_label.setStyleSheet("font-size: 18px; font-weight: bold; margin-bottom: 20px;")
        layout.addWidget(title_label)
        
        # Başlangıç ayarları
        startup_group = QGroupBox("Başlangıç")
        startup_layout = QVBoxLayout(startup_group)
        
        self.auto_login = QCheckBox("Otomatik giriş")
        startup_layout.addWidget(self.auto_login)
        
        self.startup_sound = QCheckBox("Başlangıç sesi")
        startup_layout.addWidget(self.startup_sound)
        
        layout.addWidget(startup_group)
        
        # Güvenlik ayarları
        security_group = QGroupBox("Güvenlik")
        security_layout = QFormLayout(security_group)
        
        self.session_timeout = QSpinBox()
        self.session_timeout.setRange(0, 1440)  # 0-24 saat
        self.session_timeout.setValue(0)
        self.session_timeout.setSuffix(" dakika (0 = sınırsız)")
        security_layout.addRow("Oturum zaman aşımı:", self.session_timeout)
        
        self.require_password = QCheckBox("Uyku modundan çıkarken şifre iste")
        security_layout.addRow("", self.require_password)
        
        layout.addWidget(security_group)
        
        # Performans ayarları
        performance_group = QGroupBox("Performans")
        performance_layout = QFormLayout(performance_group)
        
        self.animations = QCheckBox("Animasyonları etkinleştir")
        self.animations.setChecked(True)
        performance_layout.addRow("", self.animations)
        
        self.transparency = QCheckBox("Saydamlık efektleri")
        self.transparency.setChecked(True)
        performance_layout.addRow("", self.transparency)
        
        self.memory_limit = QSpinBox()
        self.memory_limit.setRange(256, 8192)
        self.memory_limit.setValue(1024)
        self.memory_limit.setSuffix(" MB")
        performance_layout.addRow("Bellek sınırı:", self.memory_limit)
        
        layout.addWidget(performance_group)
        
        # Sistem bilgileri
        info_group = QGroupBox("Sistem Bilgileri")
        info_layout = QFormLayout(info_group)
        
        info_layout.addRow("İşletim Sistemi:", QLabel("PyCloud OS 0.9.0-dev"))
        info_layout.addRow("Python Sürümü:", QLabel(f"{sys.version.split()[0]}"))
        info_layout.addRow("PyQt Sürümü:", QLabel("6.x"))
        
        layout.addWidget(info_group)
        
        layout.addStretch()
        
        # Sinyal bağlantıları
        self.auto_login.toggled.connect(self.on_settings_changed)
        self.startup_sound.toggled.connect(self.on_settings_changed)
        self.session_timeout.valueChanged.connect(self.on_settings_changed)
        self.require_password.toggled.connect(self.on_settings_changed)
        self.animations.toggled.connect(self.on_settings_changed)
        self.transparency.toggled.connect(self.on_settings_changed)
        self.memory_limit.valueChanged.connect(self.on_settings_changed)
    
    def on_settings_changed(self):
        """Ayar değişti"""
        settings = {
            "auto_login": self.auto_login.isChecked(),
            "startup_sound": self.startup_sound.isChecked(),
            "session_timeout": self.session_timeout.value(),
            "require_password": self.require_password.isChecked(),
            "animations": self.animations.isChecked(),
            "transparency": self.transparency.isChecked(),
            "memory_limit": self.memory_limit.value()
        }
        
        self.settings = settings
        self.settings_changed.emit("system", settings)
    
    def update_ui(self):
        """UI'yi ayarlara göre güncelle"""
        self.auto_login.setChecked(self.settings.get("auto_login", False))
        self.startup_sound.setChecked(self.settings.get("startup_sound", True))
        self.session_timeout.setValue(self.settings.get("session_timeout", 0))
        self.require_password.setChecked(self.settings.get("require_password", True))
        self.animations.setChecked(self.settings.get("animations", True))
        self.transparency.setChecked(self.settings.get("transparency", True))
        self.memory_limit.setValue(self.settings.get("memory_limit", 1024))

class NotificationsPage(SettingsPage):
    """Bildirim ayarları sayfası"""
    
    def __init__(self):
        super().__init__("Bildirimler", "🔔")
    
    def setup_ui(self):
        """UI kurulumu"""
        layout = QVBoxLayout(self)
        
        # Başlık
        title_label = QLabel(f"{self.icon} {self.title}")
        title_label.setStyleSheet("font-size: 18px; font-weight: bold; margin-bottom: 20px;")
        layout.addWidget(title_label)
        
        # Genel ayarlar
        general_group = QGroupBox("Genel")
        general_layout = QVBoxLayout(general_group)
        
        self.enable_notifications = QCheckBox("Bildirimleri etkinleştir")
        self.enable_notifications.setChecked(True)
        general_layout.addWidget(self.enable_notifications)
        
        self.notification_sound = QCheckBox("Bildirim sesi")
        self.notification_sound.setChecked(True)
        general_layout.addWidget(self.notification_sound)
        
        self.show_preview = QCheckBox("Bildirim önizlemesi göster")
        self.show_preview.setChecked(True)
        general_layout.addWidget(self.show_preview)
        
        layout.addWidget(general_group)
        
        # Görünüm ayarları
        display_group = QGroupBox("Görünüm")
        display_layout = QFormLayout(display_group)
        
        self.notification_position = QComboBox()
        self.notification_position.addItems(["Sağ Üst", "Sol Üst", "Sağ Alt", "Sol Alt", "Merkez"])
        display_layout.addRow("Konum:", self.notification_position)
        
        self.notification_duration = QSpinBox()
        self.notification_duration.setRange(1, 30)
        self.notification_duration.setValue(5)
        self.notification_duration.setSuffix(" saniye")
        display_layout.addRow("Süre:", self.notification_duration)
        
        self.max_notifications = QSpinBox()
        self.max_notifications.setRange(1, 10)
        self.max_notifications.setValue(3)
        display_layout.addRow("Maksimum sayı:", self.max_notifications)
        
        layout.addWidget(display_group)
        
        # Uygulama bildirimleri
        apps_group = QGroupBox("Uygulama Bildirimleri")
        apps_layout = QVBoxLayout(apps_group)
        
        # Uygulama listesi (örnek)
        app_list = QListWidget()
        app_items = [
            ("Sistem", True),
            ("Güvenlik", True),
            ("Güncellemeler", True),
            ("Clapp", False),
            ("Cloud Files", False)
        ]
        
        for app_name, enabled in app_items:
            item = QListWidgetItem(app_name)
            item.setFlags(item.flags() | Qt.ItemFlag.ItemIsUserCheckable)
            item.setCheckState(Qt.CheckState.Checked if enabled else Qt.CheckState.Unchecked)
            app_list.addItem(item)
        
        apps_layout.addWidget(app_list)
        layout.addWidget(apps_group)
        
        layout.addStretch()
        
        # Sinyal bağlantıları
        self.enable_notifications.toggled.connect(self.on_settings_changed)
        self.notification_sound.toggled.connect(self.on_settings_changed)
        self.show_preview.toggled.connect(self.on_settings_changed)
        self.notification_position.currentTextChanged.connect(self.on_settings_changed)
        self.notification_duration.valueChanged.connect(self.on_settings_changed)
        self.max_notifications.valueChanged.connect(self.on_settings_changed)
    
    def on_settings_changed(self):
        """Ayar değişti"""
        settings = {
            "enable_notifications": self.enable_notifications.isChecked(),
            "notification_sound": self.notification_sound.isChecked(),
            "show_preview": self.show_preview.isChecked(),
            "notification_position": self.notification_position.currentText().lower().replace(" ", "_"),
            "notification_duration": self.notification_duration.value(),
            "max_notifications": self.max_notifications.value()
        }
        
        self.settings = settings
        self.settings_changed.emit("notifications", settings)
    
    def update_ui(self):
        """UI'yi ayarlara göre güncelle"""
        self.enable_notifications.setChecked(self.settings.get("enable_notifications", True))
        self.notification_sound.setChecked(self.settings.get("notification_sound", True))
        self.show_preview.setChecked(self.settings.get("show_preview", True))
        
        position = self.settings.get("notification_position", "sağ_üst").replace("_", " ").title()
        index = self.notification_position.findText(position)
        if index >= 0:
            self.notification_position.setCurrentIndex(index)
        
        self.notification_duration.setValue(self.settings.get("notification_duration", 5))
        self.max_notifications.setValue(self.settings.get("max_notifications", 3))

class CloudSettings(QMainWindow):
    """Ana ayarlar uygulaması"""
    
    def __init__(self, kernel=None):
        super().__init__()
        self.kernel = kernel
        self.logger = logging.getLogger("CloudSettings")
        
        # Ayar dosyası
        self.settings_file = Path("system/config/ui_settings.json")
        self.settings_file.parent.mkdir(parents=True, exist_ok=True)
        
        # Ayarlar
        self.all_settings = {}
        
        # Sayfalar
        self.pages = {}
        
        # UI kurulumu
        self.setup_ui()
        self.load_settings()
        
        # İlk sayfa
        self.show_page("appearance")
    
    def setup_ui(self):
        """UI kurulumu"""
        self.setWindowTitle("PyCloud Ayarları")
        self.setGeometry(100, 100, 800, 600)
        
        # Ana widget
        central_widget = QWidget()
        self.setCentralWidget(central_widget)
        
        # Ana layout
        main_layout = QHBoxLayout(central_widget)
        main_layout.setSpacing(0)
        
        # Sol panel (kategori listesi)
        self.setup_sidebar()
        main_layout.addWidget(self.sidebar)
        
        # Sağ panel (ayar sayfaları)
        self.setup_content_area()
        main_layout.addWidget(self.content_area, 1)
        
        # Durum çubuğu
        self.setup_statusbar()
    
    def setup_sidebar(self):
        """Sol panel kurulumu"""
        self.sidebar = QWidget()
        self.sidebar.setFixedWidth(200)
        self.sidebar.setStyleSheet("""
            QWidget {
                background-color: #f5f5f5;
                border-right: 1px solid #ddd;
            }
        """)
        
        layout = QVBoxLayout(self.sidebar)
        layout.setContentsMargins(10, 10, 10, 10)
        
        # Başlık
        title_label = QLabel("Ayarlar")
        title_label.setStyleSheet("font-weight: bold; font-size: 16px; margin-bottom: 20px;")
        layout.addWidget(title_label)
        
        # Kategori listesi
        self.category_list = QListWidget()
        self.category_list.setStyleSheet("""
            QListWidget {
                border: none;
                background-color: transparent;
            }
            QListWidget::item {
                padding: 12px 8px;
                border-radius: 6px;
                margin: 2px 0;
                font-size: 14px;
            }
            QListWidget::item:selected {
                background-color: #2196f3;
                color: white;
            }
            QListWidget::item:hover {
                background-color: #e3f2fd;
            }
        """)
        
        # Kategoriler
        categories = [
            ("appearance", "🎨 Görünüm"),
            ("system", "💻 Sistem"),
            ("notifications", "🔔 Bildirimler")
        ]
        
        for category_id, category_name in categories:
            item = QListWidgetItem(category_name)
            item.setData(Qt.ItemDataRole.UserRole, category_id)
            self.category_list.addItem(item)
        
        self.category_list.setCurrentRow(0)
        self.category_list.currentItemChanged.connect(self.on_category_changed)
        
        layout.addWidget(self.category_list)
        layout.addStretch()
        
        # Alt butonlar
        buttons_layout = QVBoxLayout()
        
        # Varsayılanlara dön
        reset_button = QPushButton("Varsayılanlara Dön")
        reset_button.setStyleSheet("""
            QPushButton {
                background-color: #ff9800;
                color: white;
                border: none;
                padding: 8px;
                border-radius: 6px;
                font-weight: bold;
            }
            QPushButton:hover {
                background-color: #f57c00;
            }
        """)
        reset_button.clicked.connect(self.reset_to_defaults)
        buttons_layout.addWidget(reset_button)
        
        # Uygula
        apply_button = QPushButton("Uygula")
        apply_button.setStyleSheet("""
            QPushButton {
                background-color: #4caf50;
                color: white;
                border: none;
                padding: 8px;
                border-radius: 6px;
                font-weight: bold;
            }
            QPushButton:hover {
                background-color: #45a049;
            }
        """)
        apply_button.clicked.connect(self.apply_settings)
        buttons_layout.addWidget(apply_button)
        
        layout.addLayout(buttons_layout)
    
    def setup_content_area(self):
        """İçerik alanı kurulumu"""
        self.content_area = QStackedWidget()
        
        # Sayfaları oluştur
        self.pages["appearance"] = AppearancePage(self.kernel)
        self.pages["system"] = SystemPage()
        self.pages["notifications"] = NotificationsPage()
        
        # Sayfaları ekle
        for page in self.pages.values():
            self.content_area.addWidget(page)
            page.settings_changed.connect(self.on_settings_changed)
    
    def setup_statusbar(self):
        """Durum çubuğu kurulumu"""
        self.status_bar = self.statusBar()
        self.status_label = QLabel("Hazır")
        self.status_bar.addWidget(self.status_label)
        
        # Sağ tarafta son kaydetme zamanı
        self.last_saved_label = QLabel("")
        self.status_bar.addPermanentWidget(self.last_saved_label)
    
    def on_category_changed(self, current, previous):
        """Kategori değişti"""
        if current:
            category_id = current.data(Qt.ItemDataRole.UserRole)
            self.show_page(category_id)
    
    def show_page(self, category_id: str):
        """Sayfa göster"""
        if category_id in self.pages:
            page = self.pages[category_id]
            self.content_area.setCurrentWidget(page)
            
            # Sayfaya ayarları yükle
            if category_id in self.all_settings:
                page.load_settings(self.all_settings[category_id])
    
    def on_settings_changed(self, category: str, settings: Dict):
        """Ayar değişti"""
        self.all_settings[category] = settings
        self.status_label.setText(f"{category.title()} ayarları değiştirildi")
    
    def load_settings(self):
        """Ayarları yükle"""
        try:
            if self.settings_file.exists():
                with open(self.settings_file, 'r', encoding='utf-8') as f:
                    self.all_settings = json.load(f)
                
                self.logger.info("Settings loaded successfully")
                self.status_label.setText("Ayarlar yüklendi")
            else:
                # Varsayılan ayarlar
                self.all_settings = self.get_default_settings()
                self.save_settings()
                
        except Exception as e:
            self.logger.error(f"Failed to load settings: {e}")
            self.all_settings = self.get_default_settings()
    
    def save_settings(self):
        """Ayarları kaydet"""
        try:
            with open(self.settings_file, 'w', encoding='utf-8') as f:
                json.dump(self.all_settings, f, indent=2, ensure_ascii=False)
            
            self.logger.info("Settings saved successfully")
            self.last_saved_label.setText(f"Son kayıt: {datetime.now().strftime('%H:%M:%S')}")
            
            # Kernel'e bildir
            if self.kernel:
                config_manager = self.kernel.get_module("config")
                if config_manager:
                    config_manager.set("ui_settings", self.all_settings)
            
        except Exception as e:
            self.logger.error(f"Failed to save settings: {e}")
            QMessageBox.critical(self, "Hata", f"Ayarlar kaydedilemedi:\n{e}")
    
    def apply_settings(self):
        """Ayarları uygula"""
        self.save_settings()
        self.status_label.setText("Ayarlar uygulandı")
        
        # Rain UI'ye bildir
        if self.kernel:
            rain_ui = self.kernel.get_module("rain_ui")
            if rain_ui:
                rain_ui.apply_settings(self.all_settings)
    
    def reset_to_defaults(self):
        """Varsayılan ayarlara dön"""
        reply = QMessageBox.question(
            self, "Varsayılanlara Dön",
            "Tüm ayarları varsayılan değerlere döndürmek istediğinizden emin misiniz?",
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No
        )
        
        if reply == QMessageBox.StandardButton.Yes:
            self.all_settings = self.get_default_settings()
            
            # Tüm sayfaları güncelle
            for category_id, page in self.pages.items():
                if category_id in self.all_settings:
                    page.load_settings(self.all_settings[category_id])
            
            self.save_settings()
            self.status_label.setText("Varsayılan ayarlar yüklendi")
    
    def get_default_settings(self) -> Dict:
        """Varsayılan ayarları al"""
        return {
            "appearance": {
                "theme": "auto",
                "wallpaper_path": "",
                "wallpaper_mode": "sığdır",
                "dock_position": "alt",
                "dock_size": 64,
                "dock_autohide": False,
                "show_clock": True,
                "show_user": True,
                "show_notifications": True
            },
            "system": {
                "auto_login": False,
                "startup_sound": True,
                "session_timeout": 0,
                "require_password": True,
                "animations": True,
                "transparency": True,
                "memory_limit": 1024
            },
            "notifications": {
                "enable_notifications": True,
                "notification_sound": True,
                "show_preview": True,
                "notification_position": "sağ_üst",
                "notification_duration": 5,
                "max_notifications": 3
            }
        }
    
    def closeEvent(self, event):
        """Pencere kapatılıyor"""
        # Değişiklikleri kaydet
        self.save_settings()
        event.accept()

def main():
    """Ana fonksiyon"""
    app = QApplication(sys.argv)
    app.setApplicationName("PyCloud Settings")
    app.setApplicationVersion("1.0.0")
    
    # Ana pencere
    window = CloudSettings()
    window.show()
    
    return app.exec()

if __name__ == "__main__":
    sys.exit(main()) 


""" brain-dead simple parser for ini-style files.
(C) Ronny Pfannschmidt, Holger Krekel -- MIT licensed
"""
from __future__ import annotations
from typing import (
    Callable,
    Iterator,
    Mapping,
    Optional,
    Tuple,
    TypeVar,
    Union,
    TYPE_CHECKING,
    NoReturn,
    NamedTuple,
    overload,
    cast,
)

import os

if TYPE_CHECKING:
    from typing import Final

__all__ = ["IniConfig", "ParseError", "COMMENTCHARS", "iscommentline"]

from .exceptions import ParseError
from . import _parse
from ._parse import COMMENTCHARS, iscommentline

_D = TypeVar("_D")
_T = TypeVar("_T")


class SectionWrapper:
    config: Final[IniConfig]
    name: Final[str]

    def __init__(self, config: IniConfig, name: str) -> None:
        self.config = config
        self.name = name

    def lineof(self, name: str) -> int | None:
        return self.config.lineof(self.name, name)

    @overload
    def get(self, key: str) -> str | None:
        ...

    @overload
    def get(
        self,
        key: str,
        convert: Callable[[str], _T],
    ) -> _T | None:
        ...

    @overload
    def get(
        self,
        key: str,
        default: None,
        convert: Callable[[str], _T],
    ) -> _T | None:
        ...

    @overload
    def get(self, key: str, default: _D, convert: None = None) -> str | _D:
        ...

    @overload
    def get(
        self,
        key: str,
        default: _D,
        convert: Callable[[str], _T],
    ) -> _T | _D:
        ...

    # TODO: investigate possible mypy bug wrt matching the passed over data
    def get(  # type: ignore [misc]
        self,
        key: str,
        default: _D | None = None,
        convert: Callable[[str], _T] | None = None,
    ) -> _D | _T | str | None:
        return self.config.get(self.name, key, convert=convert, default=default)

    def __getitem__(self, key: str) -> str:
        return self.config.sections[self.name][key]

    def __iter__(self) -> Iterator[str]:
        section: Mapping[str, str] = self.config.sections.get(self.name, {})

        def lineof(key: str) -> int:
            return self.config.lineof(self.name, key)  # type: ignore[return-value]

        yield from sorted(section, key=lineof)

    def items(self) -> Iterator[tuple[str, str]]:
        for name in self:
            yield name, self[name]


class IniConfig:
    path: Final[str]
    sections: Final[Mapping[str, Mapping[str, str]]]

    def __init__(
        self,
        path: str | os.PathLike[str],
        data: str | None = None,
        encoding: str = "utf-8",
    ) -> None:
        self.path = os.fspath(path)
        if data is None:
            with open(self.path, encoding=encoding) as fp:
                data = fp.read()

        tokens = _parse.parse_lines(self.path, data.splitlines(True))

        self._sources = {}
        sections_data: dict[str, dict[str, str]]
        self.sections = sections_data = {}

        for lineno, section, name, value in tokens:
            if section is None:
                raise ParseError(self.path, lineno, "no section header defined")
            self._sources[section, name] = lineno
            if name is None:
                if section in self.sections:
                    raise ParseError(
                        self.path, lineno, f"duplicate section {section!r}"
                    )
                sections_data[section] = {}
            else:
                if name in self.sections[section]:
                    raise ParseError(self.path, lineno, f"duplicate name {name!r}")
                assert value is not None
                sections_data[section][name] = value

    def lineof(self, section: str, name: str | None = None) -> int | None:
        lineno = self._sources.get((section, name))
        return None if lineno is None else lineno + 1

    @overload
    def get(
        self,
        section: str,
        name: str,
    ) -> str | None:
        ...

    @overload
    def get(
        self,
        section: str,
        name: str,
        convert: Callable[[str], _T],
    ) -> _T | None:
        ...

    @overload
    def get(
        self,
        section: str,
        name: str,
        default: None,
        convert: Callable[[str], _T],
    ) -> _T | None:
        ...

    @overload
    def get(
        self, section: str, name: str, default: _D, convert: None = None
    ) -> str | _D:
        ...

    @overload
    def get(
        self,
        section: str,
        name: str,
        default: _D,
        convert: Callable[[str], _T],
    ) -> _T | _D:
        ...

    def get(  # type: ignore
        self,
        section: str,
        name: str,
        default: _D | None = None,
        convert: Callable[[str], _T] | None = None,
    ) -> _D | _T | str | None:
        try:
            value: str = self.sections[section][name]
        except KeyError:
            return default
        else:
            if convert is not None:
                return convert(value)
            else:
                return value

    def __getitem__(self, name: str) -> SectionWrapper:
        if name not in self.sections:
            raise KeyError(name)
        return SectionWrapper(self, name)

    def __iter__(self) -> Iterator[SectionWrapper]:
        for name in sorted(self.sections, key=self.lineof):  # type: ignore
            yield SectionWrapper(self, name)

    def __contains__(self, arg: str) -> bool:
        return arg in self.sections



from __future__ import annotations

from collections.abc import Iterable
from collections.abc import Sequence
import os
from pathlib import Path
import sys
from typing import TYPE_CHECKING

import iniconfig

from .exceptions import UsageError
from _pytest.outcomes import fail
from _pytest.pathlib import absolutepath
from _pytest.pathlib import commonpath
from _pytest.pathlib import safe_exists


if TYPE_CHECKING:
    from typing import Union

    from typing_extensions import TypeAlias

    # Even though TOML supports richer data types, all values are converted to str/list[str] during
    # parsing to maintain compatibility with the rest of the configuration system.
    ConfigDict: TypeAlias = dict[str, Union[str, list[str]]]


def _parse_ini_config(path: Path) -> iniconfig.IniConfig:
    """Parse the given generic '.ini' file using legacy IniConfig parser, returning
    the parsed object.

    Raise UsageError if the file cannot be parsed.
    """
    try:
        return iniconfig.IniConfig(str(path))
    except iniconfig.ParseError as exc:
        raise UsageError(str(exc)) from exc


def load_config_dict_from_file(
    filepath: Path,
) -> ConfigDict | None:
    """Load pytest configuration from the given file path, if supported.

    Return None if the file does not contain valid pytest configuration.
    """
    # Configuration from ini files are obtained from the [pytest] section, if present.
    if filepath.suffix == ".ini":
        iniconfig = _parse_ini_config(filepath)

        if "pytest" in iniconfig:
            return dict(iniconfig["pytest"].items())
        else:
            # "pytest.ini" files are always the source of configuration, even if empty.
            if filepath.name == "pytest.ini":
                return {}

    # '.cfg' files are considered if they contain a "[tool:pytest]" section.
    elif filepath.suffix == ".cfg":
        iniconfig = _parse_ini_config(filepath)

        if "tool:pytest" in iniconfig.sections:
            return dict(iniconfig["tool:pytest"].items())
        elif "pytest" in iniconfig.sections:
            # If a setup.cfg contains a "[pytest]" section, we raise a failure to indicate users that
            # plain "[pytest]" sections in setup.cfg files is no longer supported (#3086).
            fail(CFG_PYTEST_SECTION.format(filename="setup.cfg"), pytrace=False)

    # '.toml' files are considered if they contain a [tool.pytest.ini_options] table.
    elif filepath.suffix == ".toml":
        if sys.version_info >= (3, 11):
            import tomllib
        else:
            import tomli as tomllib

        toml_text = filepath.read_text(encoding="utf-8")
        try:
            config = tomllib.loads(toml_text)
        except tomllib.TOMLDecodeError as exc:
            raise UsageError(f"{filepath}: {exc}") from exc

        result = config.get("tool", {}).get("pytest", {}).get("ini_options", None)
        if result is not None:
            # TOML supports richer data types than ini files (strings, arrays, floats, ints, etc),
            # however we need to convert all scalar values to str for compatibility with the rest
            # of the configuration system, which expects strings only.
            def make_scalar(v: object) -> str | list[str]:
                return v if isinstance(v, list) else str(v)

            return {k: make_scalar(v) for k, v in result.items()}

    return None


def locate_config(
    invocation_dir: Path,
    args: Iterable[Path],
) -> tuple[Path | None, Path | None, ConfigDict]:
    """Search in the list of arguments for a valid ini-file for pytest,
    and return a tuple of (rootdir, inifile, cfg-dict)."""
    config_names = [
        "pytest.ini",
        ".pytest.ini",
        "pyproject.toml",
        "tox.ini",
        "setup.cfg",
    ]
    args = [x for x in args if not str(x).startswith("-")]
    if not args:
        args = [invocation_dir]
    found_pyproject_toml: Path | None = None
    for arg in args:
        argpath = absolutepath(arg)
        for base in (argpath, *argpath.parents):
            for config_name in config_names:
                p = base / config_name
                if p.is_file():
                    if p.name == "pyproject.toml" and found_pyproject_toml is None:
                        found_pyproject_toml = p
                    ini_config = load_config_dict_from_file(p)
                    if ini_config is not None:
                        return base, p, ini_config
    if found_pyproject_toml is not None:
        return found_pyproject_toml.parent, found_pyproject_toml, {}
    return None, None, {}


def get_common_ancestor(
    invocation_dir: Path,
    paths: Iterable[Path],
) -> Path:
    common_ancestor: Path | None = None
    for path in paths:
        if not path.exists():
            continue
        if common_ancestor is None:
            common_ancestor = path
        else:
            if common_ancestor in path.parents or path == common_ancestor:
                continue
            elif path in common_ancestor.parents:
                common_ancestor = path
            else:
                shared = commonpath(path, common_ancestor)
                if shared is not None:
                    common_ancestor = shared
    if common_ancestor is None:
        common_ancestor = invocation_dir
    elif common_ancestor.is_file():
        common_ancestor = common_ancestor.parent
    return common_ancestor


def get_dirs_from_args(args: Iterable[str]) -> list[Path]:
    def is_option(x: str) -> bool:
        return x.startswith("-")

    def get_file_part_from_node_id(x: str) -> str:
        return x.split("::")[0]

    def get_dir_from_path(path: Path) -> Path:
        if path.is_dir():
            return path
        return path.parent

    # These look like paths but may not exist
    possible_paths = (
        absolutepath(get_file_part_from_node_id(arg))
        for arg in args
        if not is_option(arg)
    )

    return [get_dir_from_path(path) for path in possible_paths if safe_exists(path)]


CFG_PYTEST_SECTION = "[pytest] section in {filename} files is no longer supported, change to [tool:pytest] instead."


def determine_setup(
    *,
    inifile: str | None,
    args: Sequence[str],
    rootdir_cmd_arg: str | None,
    invocation_dir: Path,
) -> tuple[Path, Path | None, ConfigDict]:
    """Determine the rootdir, inifile and ini configuration values from the
    command line arguments.

    :param inifile:
        The `--inifile` command line argument, if given.
    :param args:
        The free command line arguments.
    :param rootdir_cmd_arg:
        The `--rootdir` command line argument, if given.
    :param invocation_dir:
        The working directory when pytest was invoked.
    """
    rootdir = None
    dirs = get_dirs_from_args(args)
    if inifile:
        inipath_ = absolutepath(inifile)
        inipath: Path | None = inipath_
        inicfg = load_config_dict_from_file(inipath_) or {}
        if rootdir_cmd_arg is None:
            rootdir = inipath_.parent
    else:
        ancestor = get_common_ancestor(invocation_dir, dirs)
        rootdir, inipath, inicfg = locate_config(invocation_dir, [ancestor])
        if rootdir is None and rootdir_cmd_arg is None:
            for possible_rootdir in (ancestor, *ancestor.parents):
                if (possible_rootdir / "setup.py").is_file():
                    rootdir = possible_rootdir
                    break
            else:
                if dirs != [ancestor]:
                    rootdir, inipath, inicfg = locate_config(invocation_dir, dirs)
                if rootdir is None:
                    rootdir = get_common_ancestor(
                        invocation_dir, [invocation_dir, ancestor]
                    )
                    if is_fs_root(rootdir):
                        rootdir = ancestor
    if rootdir_cmd_arg:
        rootdir = absolutepath(os.path.expandvars(rootdir_cmd_arg))
        if not rootdir.is_dir():
            raise UsageError(
                f"Directory '{rootdir}' not found. Check your '--rootdir' option."
            )
    assert rootdir is not None
    return rootdir, inipath, inicfg or {}


def is_fs_root(p: Path) -> bool:
    r"""
    Return True if the given path is pointing to the root of the
    file system ("/" on Unix and "C:\\" on Windows for example).
    """
    return os.path.splitdrive(str(p))[1] == os.sep
